# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any
from typing import Dict
from typing import List

from anemoi.inference.config.run import RunConfiguration
from anemoi.inference.forcings import CoupledForcings
from anemoi.inference.forcings import Forcings
from anemoi.inference.output import Output
from anemoi.inference.runners.default import DefaultRunner
from anemoi.inference.runners.testing import TestingMixing
from anemoi.inference.transport import Coupling
from anemoi.inference.transport import Transport
from anemoi.inference.types import Date
from anemoi.inference.types import FloatArray
from anemoi.inference.types import State

from ..task import Task
from . import task_registry

LOG = logging.getLogger(__name__)


class CouplingForcings(CoupledForcings):
    """Just to have a different __repr__."""


class CoupledRunner(DefaultRunner):
    """Runner for coupled models.

    This class handles the initialization and running of coupled models
    using the provided configuration and input.
    """

    def __init__(self, config: Dict[str, Any], coupled_input: "CoupledInput") -> None:
        """Initialize the CoupledRunner.

        Parameters
        ----------
        config : dict
            Configuration dictionary.
        coupled_input : CoupledInput
            Coupled input instance.
        """
        super().__init__(config)
        self.coupled_input = coupled_input

    def input_state_hook(self, input_state: State) -> None:
        """Hook used by coupled runners to send the input state."""
        self.coupled_input.initial_state(Output.reduce(input_state))

    def create_dynamic_coupled_forcings(self, variables: List[str], mask: Any) -> List[CoupledForcings]:
        """Create dynamic coupled forcings.

        Parameters
        ----------
        variables : list of str
            List of variable names.
        mask : Any
            Mask to apply to the variables.

        Returns
        -------
        list of CoupledForcings
            List of coupled forcings.
        """
        result = CouplingForcings(self, self.coupled_input, variables, mask)
        return [result]

    def initial_dynamic_forcings_inputs(self, dynamic_forcings_inputs: list[Forcings]) -> list[Forcings]:
        """Modify the dynamic forcings inputs for the first step.

        Parameters
        ----------
        dynamic_forcings_inputs : list of Forcings
            The dynamic forcings inputs.

        Returns
        -------
        list[Forcings]
            The modified dynamic forcings inputs.
        """
        # For the initial state we need to load the forcings
        # from the default input.
        result = []
        for f in dynamic_forcings_inputs:
            if isinstance(f, CoupledForcings):
                result.extend(super().create_dynamic_coupled_forcings(f.variables, f.mask))
            else:
                result.append(f)
        return result

    def create_dynamic_forcings_inputs(self, input_state: State) -> list[Forcings]:
        """Create dynamic forcings inputs.

        Parameters
        ----------
        input_state : State
            The input state.

        Returns
        -------
        list[Forcings]
            The created dynamic forcings inputs.
        """
        forcings = super().create_dynamic_forcings_inputs(input_state)
        result = []
        for f in forcings:
            if isinstance(f, CoupledForcings):
                # Substituting the CoupledForcings input with the coupled input
                # TODO: review this
                f = CouplingForcings(self, self.coupled_input, f.variables, f.mask)

            result.append(f)

        return result


class CoupledInput:
    """Input handler for coupled models.

    This class manages the input data and state for coupled models,
    including loading and initializing forcings.
    """

    trace_name = "coupled"

    def __init__(self, task: Task, transport: Any, couplings: List[Coupling]) -> None:
        """Initialize the CoupledInput.

        Parameters
        ----------
        task : Task
            Task instance.
        transport : Any
            Transport instance.
        couplings : list of Coupling
            List of couplings.
        """
        self.task = task
        self.transport = transport
        self.couplings = couplings
        self.constants: Dict[str, FloatArray] = {}
        self.tag = 0

    def load_forcings_state(self, *, variables: List[str], dates: List[Date], current_state: State) -> State:
        """Load the forcings state.

        Parameters
        ----------
        variables : list of str
            List of variable names.
        dates : list of Date
            List of dates.
        current_state : State
            Current state dictionary.

        Returns
        -------
        State
            Updated state dictionary.
        """
        LOG.info("Adding dynamic forcings %s %s", variables, dates)
        state = dict(variables=variables, date=dates)

        for c in self.couplings:
            c.apply(
                self.task,
                self.transport,
                input_state=current_state,
                output_state=state,
                constants=self.constants,
                tag=self.tag,
            )

        for f, v in state["fields"].items():
            assert len(v.shape) == 1, (f, v.shape)

        assert state["date"] == dates, (state["date"], dates)

        self.tag += 1

        return state

    def initial_state(self, state: State) -> None:
        """Initialize the state.

        Parameters
        ----------
        state : State
            State dictionary.
        """
        # We want to copy the constants that may be requested by the other tasks
        # For now, we keep it simple and just copy the whole state
        self.constants = state["fields"].copy()


class TestCoupledRunner(TestingMixing, CoupledRunner):
    """Runner for testing coupled models."""

    def __init__(self, config: Dict[str, Any], coupled_input: "CoupledInput") -> None:
        """Initialize the TestCoupledRunner.

        Parameters
        ----------
        config : dict
            Configuration dictionary.
        coupled_input : CoupledInput
            Coupled input instance.
        """

        super().__init__(config, coupled_input)


@task_registry.register("runner")
class RunnerTask(Task):
    """Task for running coupled models.

    This task initializes and runs coupled models using the provided
    configuration, overrides, and global configuration.
    """

    def __init__(
        self, name: str, config: Dict[str, Any], overrides: Dict[str, Any] = {}, global_config: Dict[str, Any] = {}
    ) -> None:
        """Initialize the RunnerTask.

        Parameters
        ----------
        name : str
            Name of the task.
        config : dict
            Configuration dictionary.
        overrides : dict, optional
            Overrides dictionary.
        global_config : dict, optional
            Global configuration dictionary.
        """
        super().__init__(name)
        LOG.info("Creating RunnerTask %s %s (%s)", self, config, global_config)
        self.config = RunConfiguration.load(config, overrides=[global_config, overrides])

    def run(self, transport: Transport) -> None:
        """Run the task.

        Parameters
        ----------
        transport : Transport
            Transport instance.
        """
        LOG.info("Running task %s", self.name)
        couplings = transport.couplings(self)

        assert self.config.runner in ("default", "testing"), self.config.runner

        coupler = CoupledInput(self, transport, couplings)

        if self.config.runner == "testing":
            runner = TestCoupledRunner(self.config, coupler)
        else:
            runner = CoupledRunner(self.config, coupler)

        runner.execute()



# (C) Copyright 2024-2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any

from anemoi.transform.filters import filter_registry

from anemoi.inference.context import Context
from anemoi.inference.decorators import main_argument
from anemoi.inference.types import State

from ..processor import Processor
from . import post_processor_registry
from .earthkit_state import unwrap_state
from .earthkit_state import wrap_state

LOG = logging.getLogger(__name__)


@post_processor_registry.register("backward_transform_filter")
@main_argument("filter")
class BackwardTransformFilter(Processor):
    """A processor that applies a backward transform filter to a given state.

    This class uses a specified filter from the filter registry to process
    the state by applying a backward transformation.

    Attributes
    ----------
    filter : Any
        The filter instance used for processing the state.
    """

    def __init__(self, context: Context, filter: str, **kwargs: Any) -> None:
        """Initialize the BackwardTransformFilter.

        Parameters
        ----------
        context : Context
            The context for the filter.
        filter : str
            The name of the filter to be used.
        **kwargs : Any
            Additional keyword arguments for the filter.
        """
        super().__init__(context)
        self.filter: Any = filter_registry.create(filter, **kwargs)

    def process(self, state: State) -> State:
        """Process the given state using the backward transform filter.

        Parameters
        ----------
        state : State
            The state to be processed.

        Returns
        -------
        State
            The processed state.
        """

        fields = self.filter.backward(wrap_state(state))

        return unwrap_state(fields, state, namer=self.context.checkpoint.default_namer())



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import logging
from typing import Any
from typing import List
from typing import Optional

from anemoi.inference.config import Configuration
from anemoi.inference.context import Context
from anemoi.inference.types import State

from ..output import ForwardOutput
from . import create_output
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("tee")
class TeeOutput(ForwardOutput):
    """TeeOutput class to manage multiple outputs."""

    def __init__(
        self,
        context: Context,
        *args: Any,
        outputs: List[Configuration],
        output_frequency: Optional[int] = None,
        write_initial_state: Optional[bool] = None,
        **kwargs: Any,
    ):
        """Initialize the TeeOutput.

        Parameters
        ----------
        context : object
            The context object.
        *args : Any
            Additional positional arguments.
        outputs : list or tuple, optional
            List of outputs to be created.
        output_frequency : int, optional
            Frequency of output.
        write_initial_state : bool, optional
            Flag to write the initial state.
        **kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)

        if outputs is None:
            outputs = args

        assert isinstance(outputs, (list, tuple)), outputs
        self.outputs = [create_output(context, output) for output in outputs]

    # We override write_initial_state and write_state
    # so users can configures each levels independently
    def write_initial_state(self, state: State) -> None:
        """Write the initial state to all outputs.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        state.setdefault("step", datetime.timedelta(0))
        for output in self.outputs:
            output.write_initial_state(state)

    def write_state(self, state: State) -> None:
        """Write the state to all outputs.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        for output in self.outputs:
            output.write_state(state)

    def write_step(self, state: State) -> None:
        """Raise NotImplementedError as TeeOutput does not support write_step.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        raise NotImplementedError("TeeOutput does not support write_step")

    def open(self, state: State) -> None:
        """Open all outputs.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        for output in self.outputs:
            output.open(state)

    def close(self) -> None:
        """Close all outputs."""
        for output in self.outputs:
            output.close()

    def __repr__(self) -> str:
        """Return a string representation of the TeeOutput.

        Returns
        -------
        str
            String representation of the TeeOutput.
        """
        return f"TeeOutput({self.outputs})"

    def print_summary(self, depth: int = 0) -> None:
        """Print the summary of all outputs.

        Parameters
        ----------
        depth : int, optional
            The depth of the summary.
        """
        super().print_summary(depth)
        for output in self.outputs:
            output.print_summary(depth + 1)



# (C) Copyright 2024-2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from datetime import timedelta
from typing import Dict
from typing import List
from typing import Optional

import numpy as np

from anemoi.inference.context import Context
from anemoi.inference.types import FloatArray
from anemoi.inference.types import State

from ..processor import Processor
from . import post_processor_registry

LOG = logging.getLogger(__name__)


@post_processor_registry.register("accumulate_from_start_of_forecast")
class Accumulate(Processor):
    """Accumulate fields from zero and return the accumulated fields.

    Parameters
    ----------
    context : Any
        The context in which the processor is running.
    accumulations : Optional[List[str]], optional
        List of fields to accumulate, by default None.
        If None, the fields are taken from the context's checkpoint.
    """

    def __init__(self, context: Context, accumulations: Optional[List[str]] = None) -> None:
        super().__init__(context)
        if accumulations is None:
            accumulations = context.checkpoint.accumulations

        self.accumulations = accumulations
        LOG.info("Accumulating fields %s", self.accumulations)

        self.accumulators: Dict[str, FloatArray] = {}
        self.step_zero = timedelta(0)

    def process(self, state: State) -> State:
        """Process the state to accumulate specified fields.

        Parameters
        ----------
        state : State
            The state containing fields to be accumulated.

        Returns
        -------
        State
            The updated state with accumulated fields.
        """
        state = state.copy()
        state.setdefault("start_steps", {})
        for accumulation in self.accumulations:
            if accumulation in state["fields"]:
                if accumulation not in self.accumulators:
                    self.accumulators[accumulation] = np.zeros_like(state["fields"][accumulation])
                self.accumulators[accumulation] += np.maximum(0, state["fields"][accumulation])
                state["fields"][accumulation] = self.accumulators[accumulation]
                state["start_steps"][accumulation] = self.step_zero

        return state

    def __repr__(self) -> str:
        """Return a string representation of the Accumulate object.

        Returns
        -------
        str
            String representation of the object.
        """
        return f"Accumulate({self.accumulations})"



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


"""Code to wrap and unwrap state dictionaries into earthkit.data field lists. So that we can pass them through anemoi-transorm filters."""

import datetime
import logging
from functools import cached_property
from typing import Any
from typing import Callable
from typing import Dict

import earthkit.data as ekd
import numpy as np
from earthkit.data.core.metadata import RawMetadata
from earthkit.data.indexing.fieldlist import SimpleFieldList

from anemoi.inference.types import FloatArray
from anemoi.inference.types import Shape
from anemoi.inference.types import State

LOG = logging.getLogger(__name__)


class StateFieldGeography:
    """Geographical information of a state field.

    Parameters
    ----------
    field : Any
        The field containing geographical data.
    """

    def __init__(self, field: Any) -> None:
        self._field = field

    @property
    def shape(self) -> Shape:
        """Tuple: Shape of the geographical field."""
        return self._field.shape


class StateFieldMetadata(RawMetadata):
    """Metadata for a state field.

    Parameters
    ----------
    field : Any
        The field containing metadata.
    """

    def __init__(self, field: Any) -> None:
        super().__init__(
            name=field.name,
            param=field.name,
            **{
                k: v
                for k, v in field.state.items()
                if isinstance(v, (str, int, float, bool, datetime.datetime, datetime.timedelta))
            },
        )
        self._field = field

    def as_namespace(self, ns: str) -> Dict[str, Any]:
        """Convert metadata to a specific namespace.

        Parameters
        ----------
        ns : str
            The namespace to convert the metadata to.

        Returns
        -------
        Dict[str, Any]
            The metadata in the specified namespace.
        """
        assert ns == "mars"
        return {k: v for k, v in self.items() if k != "name"}

    @property
    def geography(self) -> StateFieldGeography:
        """StateFieldGeography: Geographical information of the field."""
        return StateFieldGeography(self._field)


class StateField(ekd.Field):
    """State field containing name, values, and state information.

    Parameters
    ----------
    name : str
        The name of the field.
    values : FloatArray
        The values of the field.
    state : Dict[str, Any]
        The state information associated with the field.
    """

    def __init__(self, name: str, values: FloatArray, state: State) -> None:
        self.name = name
        self.__values = values
        self.state = state

    def _values(self, dtype: np.dtype) -> FloatArray:
        """Get the values of the field with a specific data type.

        Parameters
        ----------
        dtype : np.dtype
            The data type to convert the values to.

        Returns
        -------
        FloatArray
            The values of the field in the specified data type.
        """
        return self.__values.astype(dtype)

    @property
    def shape(self) -> Shape:
        """Tuple: Shape of the field."""
        return self.__values.shape

    @cached_property
    def _metadata(self) -> StateFieldMetadata:
        """StateFieldMetadata: Metadata of the field."""
        return StateFieldMetadata(self)

    def __repr__(self) -> str:
        """Str: String representation of the StateField."""
        return f"{self.__class__.__name__ }({self._metadata})"


def wrap_state(state: State) -> ekd.FieldList:
    """Transform a state dictionary into an earthkit.data field list.

    Parameters
    ----------
    state : Dict[str, Any]
        The state dictionary to be transformed.

    Returns
    -------
    ekd.FieldList
        The transformed field list.
    """
    assert isinstance(state["date"], datetime.datetime)  # Only works on single dates for now
    fields = [StateField(k, v, state) for k, v in state["fields"].items()]
    return SimpleFieldList(fields)


def unwrap_state(fields: ekd.FieldList, state: State, namer: Callable) -> State:
    """Transform a earthkit.data field list into a state dictionary.

    Parameters
    ----------
    fields : ekd.FieldList
        The field list to be transformed.
    state : State
        The original state dictionary.
    namer : Callable
        A function to generate new field names.

    Returns
    -------
    Dict[str, Any]
        The transformed state dictionary.
    """
    new_fields = {}

    # namer(field: ekd.Field, metadata: Dict[str, Any]) -> str:

    for n in fields:
        name = namer(n, n.metadata())
        new_fields[name] = n.to_numpy(flatten=True)

    state = state.copy()
    state["fields"] = new_fields

    return state



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os
from typing import Optional

import numpy as np

from anemoi.inference.context import Context
from anemoi.inference.types import State

from ..decorators import main_argument
from ..output import Output
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("raw")
@main_argument("path")
class RawOutput(Output):
    """Raw output class."""

    def __init__(
        self,
        context: Context,
        path: str,
        template: str = "{date}.npz",
        strftime: str = "%Y%m%d%H%M%S",
        output_frequency: Optional[int] = None,
        write_initial_state: Optional[bool] = None,
    ) -> None:
        """Initialize the RawOutput class.

        Parameters
        ----------
        context : dict
            The context.
        path : str
            The path to save the raw output.
        template : str, optional
            The template for filenames, by default "{date}.npz".
        strftime : str, optional
            The date format string, by default "%Y%m%d%H%M%S".
        output_frequency : int, optional
            The frequency of output, by default None.
        write_initial_state : bool, optional
            Whether to write the initial state, by default None.
        """
        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)
        self.path = path
        self.template = template
        self.strftime = strftime

    def __repr__(self) -> str:
        """Return a string representation of the RawOutput object.

        Returns
        -------
        str
            String representation of the RawOutput object.
        """
        return f"RawOutput({self.path})"

    def write_step(self, state: State) -> None:
        """Write the state to a compressed .npz file.

        Parameters
        ----------
        state : State
            The state to be written.
        """
        os.makedirs(self.path, exist_ok=True)
        date = state["date"].strftime(self.strftime)
        fn_state = f"{self.path}/{self.template.format(date=date)}"
        restate = {f"field_{key}": val for key, val in state["fields"].items()}
        for key in ["date"]:
            restate[key] = np.array(state[key], dtype=str)
        for key in ["latitudes", "longitudes"]:
            restate[key] = np.array(state[key])
        np.savez_compressed(fn_state, **restate)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os
import threading
from typing import Optional

import numpy as np

from anemoi.inference.context import Context
from anemoi.inference.types import State

from ..decorators import main_argument
from ..output import Output
from . import output_registry

LOG = logging.getLogger(__name__)


# In case HDF5 was not compiled with thread safety on
LOCK = threading.RLock()


@output_registry.register("netcdf")
@main_argument("path")
class NetCDFOutput(Output):
    """NetCDF output class.

    Parameters
    ----------
    context : dict
        The context dictionary.
    path : str
        The path to save the NetCDF file.
    output_frequency : int, optional
        The frequency of output, by default None.
    write_initial_state : bool, optional
        Whether to write the initial state, by default None.
    """

    def __init__(
        self,
        context: Context,
        path: str,
        output_frequency: Optional[int] = None,
        write_initial_state: Optional[bool] = None,
    ) -> None:
        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)

        from netCDF4 import Dataset

        self.path = path
        self.ncfile: Optional[Dataset] = None
        self.float_size = "f4"

    def __repr__(self) -> str:
        """Return a string representation of the NetCDFOutput object."""
        return f"NetCDFOutput({self.path})"

    def open(self, state: State) -> None:
        """Open the NetCDF file and initialize dimensions and variables.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        from netCDF4 import Dataset

        with LOCK:
            if self.ncfile is not None:
                return

        # If the file exists, we may get a 'Permission denied' error
        if os.path.exists(self.path):
            os.remove(self.path)

        with LOCK:
            self.ncfile = Dataset(self.path, "w", format="NETCDF4")

        compression = {}  # dict(zlib=False, complevel=0)

        values = len(state["latitudes"])

        time = 0
        self.reference_date = state["date"]
        if (time_step := getattr(self.context, "time_step", None)) and (
            lead_time := getattr(self.context, "lead_time", None)
        ):
            time = lead_time // time_step
        if reference_date := getattr(self.context, "reference_date", None):
            self.reference_date = reference_date

        with LOCK:
            self.values_dim = self.ncfile.createDimension("values", values)
            self.time_dim = self.ncfile.createDimension("time", time)
            self.time_var = self.ncfile.createVariable("time", "i4", ("time",), **compression)

            self.time_var.units = "seconds since {0}".format(self.reference_date)
            self.time_var.long_name = "time"
            self.time_var.calendar = "gregorian"

        latitudes = state["latitudes"]
        with LOCK:
            self.latitude_var = self.ncfile.createVariable("latitude", self.float_size, ("values",), **compression)
            self.latitude_var.units = "degrees_north"
            self.latitude_var.long_name = "latitude"

        longitudes = state["longitudes"]
        with LOCK:
            self.longitude_var = self.ncfile.createVariable("longitude", self.float_size, ("values",), **compression)
            self.longitude_var.units = "degrees_east"
            self.longitude_var.long_name = "longitude"

        self.latitude_var[:] = latitudes
        self.longitude_var[:] = longitudes

        self.vars = {}
        self.ensure_variables(state)

        self.n = 0

    def ensure_variables(self, state: State) -> None:
        """Ensure that all variables are created in the NetCDF file.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        values = len(state["latitudes"])

        compression = {}  # dict(zlib=False, complevel=0)

        for name in state["fields"].keys():
            if name in self.vars:
                continue
            chunksizes = (1, values)

            while np.prod(chunksizes) > 1000000:
                chunksizes = tuple(int(np.ceil(x / 2)) for x in chunksizes)

            with LOCK:
                self.vars[name] = self.ncfile.createVariable(
                    name,
                    self.float_size,
                    ("time", "values"),
                    chunksizes=chunksizes,
                    **compression,
                )

    def write_step(self, state: State) -> None:
        """Write the state.

        Parameters
        ----------
        state : State
            The state dictionary.
        """

        self.ensure_variables(state)

        step = state["date"] - self.reference_date
        self.time_var[self.n] = step.total_seconds()

        for name, value in state["fields"].items():
            with LOCK:
                self.vars[name][self.n] = value

        self.n += 1

    def close(self) -> None:
        """Close the NetCDF file."""
        if self.ncfile is not None:
            with LOCK:
                self.ncfile.close()
            self.ncfile = None



# (C) Copyright 2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Union

from numpy.typing import NDArray

"""A collection of types used in the inference module.
Some of these type could be moved to anemoi.utils.types or anemoi.transform.types.
"""

State = Dict[str, Any]
"""A dictionary that represents the state of a model."""

DataRequest = Dict[str, Any]
"""A dictionary that represent a data request, like MARS, CDS, OpenData, ..."""

Date = Union[str, datetime.datetime, int]
"""A date can be a string, a datetime object or an integer. It will always be converted to a datetime object."""

IntArray = NDArray[Any]
"""A numpy array of integers."""

FloatArray = NDArray[Any]
"""A numpy array of floats."""

BoolArray = NDArray[Any]
"""A numpy array of booleans."""

Shape = Tuple[int, ...]
"""A tuple of integers representing the shape of an array."""



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import socket
import time
from contextlib import contextmanager
from typing import Generator

import torch

LOG = logging.getLogger(__name__)


@contextmanager
def ProfilingLabel(label: str, use_profiler: bool) -> Generator[None, None, None]:
    """Add label to function so that the profiler can recognize it, only if the use_profiler option is True.

    Parameters
    ----------
    label : str
        Name or description to identify the function.
    use_profiler : bool
        Wrap the function with the label if True, otherwise just execute the function as it is.

    Returns
    -------
    Generator[None, None, None]
        Yields to the caller.
    """
    if use_profiler:
        with torch.autograd.profiler.record_function(label):
            torch.cuda.nvtx.range_push(label)
            yield
            torch.cuda.nvtx.range_pop()
    else:
        yield


@contextmanager
def ProfilingRunner(use_profiler: bool) -> Generator[None, None, None]:
    """Perform time and memory usage profiles of the wrapped code.

    Parameters
    ----------
    use_profiler : bool
        Whether to profile the wrapped code (True) or not (False).

    Returns
    -------
    Generator[None, None, None]
        Yields to the caller.
    """
    dirname = f"profiling-output/{socket.gethostname()}-{int(time.time())}"
    if use_profiler:
        torch.cuda.memory._record_memory_history(max_entries=100000)
        activities = [torch.profiler.ProfilerActivity.CPU]
        if torch.cuda.is_available():
            activities.append(torch.profiler.ProfilerActivity.CUDA)
        with torch.profiler.profile(
            profile_memory=True,
            record_shapes=True,
            activities=activities,
            with_flops=True,
            on_trace_ready=torch.profiler.tensorboard_trace_handler(dirname),
        ) as prof:
            yield
        try:
            torch.cuda.memory._dump_snapshot(f"{dirname}/memory_snapshot.pickle")
        except Exception as e:
            LOG.error(f"Failed to capture memory snapshot {e}")
        torch.cuda.memory._record_memory_history(enabled=None)
        row_limit = 10
        LOG.info(
            f"Top {row_limit} kernels by runtime on CPU:\n {prof.key_averages().table(sort_by='self_cpu_time_total', row_limit=row_limit)}"
        )
        LOG.info(
            f"Top {row_limit} kernels by runtime on CUDA:\n {prof.key_averages().table(sort_by='self_cuda_time_total', row_limit=row_limit)}"
        )
        LOG.info("Memory summary \n%s", torch.cuda.memory_summary())
        LOG.info(
            f"Memory snapshot and trace file stored to '{dirname}'. To view the memory snapshot, upload the pickle file to 'https://pytorch.org/memory_viz'. To view the trace file, see 'https://pytorch.org/tutorials/intermediate/tensorboard_profiler_tutorial.html#use-tensorboard-to-view-results-and-analyze-model-performance'"
        )
    else:
        yield



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

import yaml

from anemoi.inference.testing import files_for_tests

# "2t", "10u", "10v", "msl", "lsm", "z", "tcc", "tp", "cos_latitudes", "insolation"
#   0,   1,     2,     3,     4,    5,    6,      7,        8,            9

SIMPLE_METADATA = {
    "config": {
        "training": {
            "multistep_input": 2,
            "precision": "16-mixed",
        },
        "data": {
            "timestep": "6h",
            "forcing": ["lsm", "z"],
            "diagnostic": ["tcc", "tp"],
        },
    },
    "data_indices": {
        "data": {
            "input": {"full": [0, 1, 2, 3, 4, 5, 8, 9], "forcing": [4, 5, 8, 9], "diagnostic": [6, 7]},
            "output": {"full": [0, 1, 2, 3, 6, 7]},
        },
        "model": {
            "input": {"full": [0, 1, 2, 3, 4, 5, 6, 7], "forcing": [4, 5, 6, 7], "prognostic": [0, 1, 2, 3]},
            "output": {"full": [0, 1, 2, 3, 4, 5], "prognostic": [0, 1, 2, 3]},
        },
    },
    "dataset": {
        "variables": ["2t", "10u", "10v", "msl", "lsm", "z", "tcc", "tp", "cos_latitude", "insolation"],
        "shape": [365, 10, 1, 40320],
        "variables_metadata": {
            "2t": {"mars": {"param": "2t", "levtype": "sfc"}},
            "10u": {"mars": {"param": "10u", "levtype": "sfc"}},
            "10v": {"mars": {"param": "10v", "levtype": "sfc"}},
            "msl": {"mars": {"param": "msl", "levtype": "sfc"}},
            "lsm": {"constant_in_time": True, "mars": {"param": "lsm", "levtype": "sfc"}},
            "z": {"constant_in_time": True, "mars": {"param": "z", "levtype": "sfc"}},
            "tcc": {"mars": {"param": "tcc", "levtype": "sfc"}},
            "tp": {"accumulated": True, "mars": {"param": "tp", "levtype": "sfc"}},
            "cos_latitude": {"computed_forcing": True, "constant_in_time": True},
            "insolation": {"computed_forcing": True, "constant_in_time": False},
        },
        "data_request": {"grid": "O96"},
    },
}


def mock_load_metadata(path: Optional[str], *, supporting_arrays: bool = True) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Load metadata from a YAML file.

    Parameters
    ----------
    path : str, optional
        The path to the checkpoint file.
    supporting_arrays : bool, optional
        Whether to include supporting arrays, by default True.

    Returns
    -------
    dict
        The loaded metadata.
    """

    if path is None:
        metadata = SIMPLE_METADATA
    else:
        if not os.path.isabs(path):
            path = files_for_tests(os.path.join("checkpoints", path))
        name, _ = os.path.splitext(path)
        for ext in (".yaml", ".json"):
            path = f"{name}{ext}"
            if os.path.exists(path):
                break

        with open(path) as f:
            metadata = yaml.safe_load(f)

    arrays: Dict[str, Any] = {}

    if supporting_arrays:
        return metadata, arrays

    return metadata


def minimum_mock_checkpoint(metadata: Dict[str, Any]) -> Dict[str, Any]:
    """Create a minimum mock checkpoint from the given metadata.

    Parameters
    ----------
    metadata : dict
        The metadata to create the mock checkpoint from.

    Returns
    -------
    dict
        The minimum mock checkpoint.
    """

    KEEP = {"dataset.variables_metadata", "data_indices.data", "data_indices.model"}

    metadata = deepcopy(metadata)

    def drop(metadata: Dict[str, Any], reference: Dict[str, Any], *path: str) -> None:
        """Recursively drop keys from metadata that are not in the reference.

        Parameters
        ----------
        metadata : dict
            The metadata to modify.
        reference : dict
            The reference metadata to compare against.
        path : str
            The current path in the metadata.
        """

        if ".".join(path) in KEEP:
            return

        for k, v in list(metadata.items()):
            if k not in reference:
                del metadata[k]
            else:
                if isinstance(v, dict):
                    drop(metadata[k], reference[k], *path, k)

    drop(metadata, SIMPLE_METADATA)

    variables_metadata = metadata["dataset"]["variables_metadata"]

    for k, v in variables_metadata.items():
        mars = v.get("mars", {})
        for key in ("class", "date", "time", "hdate", "domain", "expver"):
            mars.pop(key, None)

    return metadata


def mock_torch_load(path: str, map_location: Any, weights_only: bool) -> Any:
    """Load a mock torch model for testing purposes.

    Parameters
    ----------
    path : str
        The path to the model file.
    map_location : Any
        The device on which to load the model.
    weights_only : bool
        Whether to load only the weights.

    Returns
    -------
    Any
        The mock torch model.
    """
    from .mock_model import MockModel

    assert weights_only is False, "Not implemented"

    metadata, arrays = mock_load_metadata(path)

    return MockModel(metadata, arrays)


class MockRunConfiguration:
    """Mock class for loading run configurations from `files_for_tests`."""

    @classmethod
    def load(cls, path: str, *args, **kwargs) -> Dict[str, Any]:
        """Load a run configuration from a given path.

        Parameters
        ----------
        path : str
            The path to the run configuration file.
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        dict
            The loaded run configuration.
        """
        from anemoi.inference.config.run import RunConfiguration

        if not os.path.isabs(path):
            path = files_for_tests(path)

        return RunConfiguration.load(path, *args, **kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from anemoi.inference.testing import fake_checkpoints


@fake_checkpoints
def test_checkpoint() -> None:
    """Test the Checkpoint class.

    Returns
    -------
    None
    """
    from anemoi.inference.checkpoint import Checkpoint

    c = Checkpoint("simple.chkpt")
    c.variables_from_input


if __name__ == "__main__":
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
import logging
from collections import defaultdict
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import earthkit.data as ekd
import numpy as np

from anemoi.inference.context import Context
from anemoi.inference.types import DataRequest
from anemoi.inference.types import FloatArray

from ..decorators import main_argument
from ..grib.encoding import GribWriter
from ..grib.encoding import check_encoding
from . import output_registry
from .grib import GribOutput

LOG = logging.getLogger(__name__)

# There is a bug with hindcasts, where these keys are not added to the 'mars' namespace
MARS_MAYBE_MISSING_KEYS = (
    "number",
    "step",
    "time",
    "date",
    "hdate",
    "type",
    "stream",
    "expver",
    "class",
    "levtype",
    "levelist",
    "param",
)


def _is_valid(mars: Dict[str, Any], keys: Dict[str, Any]) -> bool:
    """Check if the mars dictionary contains valid keys.

    Parameters
    ----------
    mars : Dict[str, Any]
        The mars dictionary.
    keys : Dict[str, Any]
        The keys dictionary.

    Returns
    -------
    bool
        True if valid, False otherwise.
    """
    if "number" in keys and "number" not in mars:
        LOG.warning("`number` is missing from mars namespace")
        return False

    if "referenceDate" in keys and "hdate" not in mars:
        LOG.warning("`hdate` is missing from mars namespace")
        return False

    return True


class ArchiveCollector:
    """Collects archive requests."""

    UNIQUE = {"date", "hdate", "time", "referenceDate", "type", "stream", "expver"}

    def __init__(self) -> None:
        self.expect = 0
        self._request = defaultdict(set)

    def add(self, field: Dict[str, Any]) -> None:
        """Add a field to the archive request.

        Parameters
        ----------
        field : Dict[str,Any]
            The field dictionary.
        """
        self.expect += 1
        for k, v in field.items():
            self._request[k].add(str(v))
            if k in self.UNIQUE:
                if len(self._request[k]) > 1:
                    raise ValueError(f"Field {field} has different values for {k}: {self._request[k]}")

    @property
    def request(self) -> DataRequest:
        """Get the archive request."""
        return {k: sorted(v) for k, v in self._request.items()}


@output_registry.register("grib")
@main_argument("path")
class GribFileOutput(GribOutput):
    """Handles grib files."""

    def __init__(
        self,
        context: Context,
        *,
        path: str,
        encoding: Optional[Dict[str, Any]] = None,
        archive_requests: Optional[Dict[str, Any]] = None,
        check_encoding: bool = True,
        templates: Optional[Union[List[str], str]] = None,
        grib1_keys: Optional[Dict[str, Any]] = None,
        grib2_keys: Optional[Dict[str, Any]] = None,
        modifiers: Optional[List[str]] = None,
        output_frequency: Optional[int] = None,
        write_initial_state: Optional[bool] = None,
        variables: Optional[List[str]] = None,
        split_output: bool = True,
    ) -> None:
        """Initialize the GribFileOutput.

        Parameters
        ----------
        context : Context
            The context.
        path : str
            The path to save the grib files.
        encoding : dict, optional
            The encoding dictionary, by default None.
        archive_requests : dict, optional
            The archive requests dictionary, by default None.
        check_encoding : bool, optional
            Whether to check encoding, by default True.
        templates : list or str, optional
            The templates list or string, by default None.
        grib1_keys : dict, optional
            The grib1 keys dictionary, by default None.
        grib2_keys : dict, optional
            The grib2 keys dictionary, by default None.
        modifiers : list, optional
            The list of modifiers, by default None.
        output_frequency : int, optional
            The frequency of output, by default None.
        write_initial_state : bool, optional
            Whether to write the initial state, by default None.
        variables : list, optional
            The list of variables, by default None.
        split_output : bool, optional
            Whether to split the output, by default True.
        """
        super().__init__(
            context,
            encoding=encoding,
            templates=templates,
            grib1_keys=grib1_keys,
            grib2_keys=grib2_keys,
            modifiers=modifiers,
            output_frequency=output_frequency,
            write_initial_state=write_initial_state,
            variables=variables,
        )
        self.path = path
        self.output = GribWriter(self.path, split_output=split_output)
        self.archiving = defaultdict(ArchiveCollector)
        self.archive_requests = archive_requests
        self.check_encoding = check_encoding
        self._namespace_bug_fix = False

    def __repr__(self) -> str:
        """Return a string representation of the GribFileOutput object."""
        return f"GribFileOutput({self.path})"

    def write_message(self, message: FloatArray, template: ekd.Field, **keys: Dict[str, Any]) -> None:
        """Write a message to the grib file.

        Parameters
        ----------
        message : FloatArray
            The message array.
        template : ekd.Field
            A ekd.Field use as a template for GRIB encoding.
        **keys : Dict[str, Any]
            Additional keys for the message.
        """
        # Make sure `name` is not in the keys, otherwise grib_encoding will fail
        if template is not None and template.metadata("name", default=None) is not None:
            # We cannot clear the metadata...
            class Dummy:
                def __init__(self, template: ekd.Field) -> None:
                    self.template = template
                    self.handle = template.handle

                def __repr__(self) -> str:
                    return f"Dummy({self.template})"

            template = Dummy(template)

        # LOG.info("Writing message to %s %s", template, keys)
        try:
            self.collect_archive_requests(
                self.output.write(
                    values=message,
                    template=template,
                    metadata=keys,
                    check_nans=self.context.allow_nans,
                ),
                template,
                **keys,
            )
        except Exception as e:
            import eccodes

            LOG.error("Error writing message to %s", self.path)
            LOG.error("eccodes: %s", eccodes.__version__)
            LOG.error("Template: %s, Keys: %s", template, keys)
            LOG.error("Exception: %s", e)
            if message is not None and np.isnan(message.data).any():
                LOG.error("Message contains NaNs (%s, %s) (allow_nans=%s)", keys, template, self.context.allow_nans)
            raise

    def collect_archive_requests(self, written: tuple, template: object, **keys: Any) -> None:
        """Collect archive requests.

        Parameters
        ----------
        written : tuple
            The written tuple.
        template : object
            The template object.
        **keys : Any
            Additional keys for the archive requests.
        """
        if not self.archive_requests and not self.check_encoding:
            return

        handle, path = written

        while True:

            if self._namespace_bug_fix:
                import eccodes
                from earthkit.data.readers.grib.codes import GribCodesHandle

                handle = GribCodesHandle(eccodes.codes_clone(handle._handle), None, None)

            mars = {k: v for k, v in handle.items("mars")}

            if _is_valid(mars, keys):
                break

            if self._namespace_bug_fix:
                raise ValueError("Namespace bug: %s" % mars)

            # Try again with the namespace bug
            LOG.warning("Namespace bug detected, trying again")
            self._namespace_bug_fix = True

        if self.check_encoding:
            check_encoding(handle, keys)

        if self.archive_requests:
            self.archiving[path].add(mars)

    def close(self) -> None:
        """Close the grib file."""
        self.output.close()

        if not self.archive_requests:
            return

        path = self.archive_requests["path"]
        extra = self.archive_requests.get("extra", {})
        patch = self.archive_requests.get("patch", {})
        indent = self.archive_requests.get("indent", None)

        def _patch(r: DataRequest) -> DataRequest:
            if self.context.config.use_grib_paramid:
                param = r.get("param", [])
                if not isinstance(param, list):
                    param = [param]

                # Check if we're using param ids already
                try:
                    float(next(iter(param)))
                except ValueError:
                    from anemoi.utils.grib import shortname_to_paramid

                    r["param"] = [shortname_to_paramid(p) for p in param]

            for k, v in patch.items():
                if v is None:
                    r.pop(k, None)
                else:
                    r[k] = v

            return r

        with open(self.archive_requests["path"], "w") as f:
            requests = []

            for path, archive in self.archiving.items():
                assert path is not None, "Path is None"
                request = dict(expect=archive.expect)
                request["source"] = path
                request.update(_patch(archive.request))
                request.update(extra)
                requests.append(request)

            json.dump(requests, f, indent=indent)
            f.write("\n")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import json
import logging
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from earthkit.data.utils.dates import to_datetime

from anemoi.inference.types import FloatArray
from anemoi.inference.types import State

from ..grib.encoding import grib_keys
from ..grib.templates.manager import TemplateManager
from ..output import Output

LOG = logging.getLogger(__name__)


class HindcastOutput:
    """Hindcast output class."""

    def __init__(self, reference_year: int) -> None:
        """Initialize the HindcastOutput object.

        Parameters
        ----------
        reference_year : int
            The reference year.
        """

        self.reference_year = reference_year

    def __call__(self, values: FloatArray, template: object, keys: dict) -> tuple:
        """Call the HindcastOutput object.

        Parameters
        ----------
        values : FloatArray
            The values array.
        template : object
            The template object.
        keys : dict
            The keys dictionary.

        Returns
        -------
        tuple
            The modified values, template, and keys.
        """

        if "date" not in keys:
            assert template.metadata("hdate", default=None) is None, template
            date = template.metadata("date")
        else:
            date = keys.pop("date")

        for k in ("date", "hdate", "eps", "productDefinitionTemplateNumber"):
            keys.pop(k, None)

        keys["edition"] = 1
        keys["localDefinitionNumber"] = 30
        keys["dataDate"] = int(to_datetime(date).strftime("%Y%m%d"))
        keys["referenceDate"] = int(to_datetime(date).replace(year=self.reference_year).strftime("%Y%m%d"))

        return values, template, keys


MODIFIERS = dict(hindcast=HindcastOutput)


def modifier_factory(modifiers: list) -> list:
    """Create a list of modifier instances.

    Parameters
    ----------
    modifiers : list
        A list of modifier configurations.

    Returns
    -------
    list
        A list of modifier instances.
    """

    if modifiers is None:
        return []

    if not isinstance(modifiers, list):
        modifiers = [modifiers]

    result = []
    for modifier in modifiers:
        assert isinstance(modifier, dict), modifier
        assert len(modifier) == 1, modifier

        klass = list(modifier.keys())[0]
        result.append(MODIFIERS[klass](**modifier[klass]))

    return result


class GribOutput(Output):
    """Handles grib."""

    def __init__(
        self,
        context: dict,
        *,
        encoding: Optional[Dict[str, Any]] = None,
        templates: Optional[Union[List[str], str]] = None,
        grib1_keys: Optional[Dict[str, Any]] = None,
        grib2_keys: Optional[Dict[str, Any]] = None,
        modifiers: Optional[List[str]] = None,
        output_frequency: Optional[int] = None,
        write_initial_state: Optional[bool] = None,
        variables: Optional[List[str]] = None,
    ) -> None:
        """Initialize the GribOutput object.

        Parameters
        ----------
        context : dict
            The context dictionary.
        encoding : dict, optional
            The encoding dictionary, by default None.
        templates : list or str, optional
            The templates list or string, by default None.
        grib1_keys : dict, optional
            The grib1 keys dictionary, by default None.
        grib2_keys : dict, optional
            The grib2 keys dictionary, by default None.
        modifiers : list, optional
            The list of modifiers, by default None.
        output_frequency : int, optional
            The frequency of output, by default None.
        write_initial_state : bool, optional
            Whether to write the initial state, by default None.
        variables : list, optional
            The list of variables, by default None.
        """

        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)
        self._first = True
        self.typed_variables = self.checkpoint.typed_variables
        self.encoding = encoding if encoding is not None else {}
        self.grib1_keys = grib1_keys if grib1_keys is not None else {}
        self.grib2_keys = grib2_keys if grib2_keys is not None else {}

        self.modifiers = modifier_factory(modifiers)
        self.variables = variables

        self.ensemble = False
        for d in (self.grib1_keys, self.grib2_keys, self.encoding):
            if "eps" in d:
                self.ensemble = d["eps"]
                break
            if d.get("type") in ("pf", "cf"):
                self.ensemble = True
                break

        self.template_manager = TemplateManager(self, templates)

        self.ensemble = False
        for d in (self.grib1_keys, self.grib2_keys, self.encoding):
            if "eps" in d:
                self.ensemble = d["eps"]
                break
            if d.get("type") in ("pf", "cf"):
                self.ensemble = True
                break

        self.template_manager = TemplateManager(self, templates)

    def write_initial_state(self, state: State) -> None:
        """Write the initial step of the state.

        Parameters
        ----------
        state : State
            The state object.
        """
        # We trust the GribInput class to provide the templates
        # matching the input state

        state = state.copy()

        self.reference_date = state["date"]
        state.setdefault("step", datetime.timedelta(0))

        out_vars = self.variables if self.variables is not None else state["fields"].keys()

        for name in out_vars:
            variable = self.typed_variables[name]

            if variable.is_computed_forcing:
                continue

            template = self.template(state, name)
            if template is None:
                # We can currently only write grib output if we have a grib input
                raise ValueError(
                    "GRIB output only works if the input is GRIB (for now). Set `write_initial_step` to `false`."
                )

        return self.write_step(state)

    def write_step(self, state: State) -> None:
        """Write a step of the state.

        Parameters
        ----------
        state : State
            The state object.
        """

        reference_date = self.reference_date or self.context.reference_date
        step = state["step"]
        previous_step = state.get("previous_step")
        start_steps = state.get("start_steps", {})

        out_vars = self.variables if self.variables is not None else state["fields"].keys()
        for name in out_vars:
            values = state["fields"][name]
            keys = {}

            variable = self.typed_variables[name]

            if variable.is_computed_forcing:
                continue

            param = variable.grib_keys.get("param", name)

            template = self.template(state, name)

            keys.update(self.encoding)

            keys = grib_keys(
                values=values,
                template=template,
                date=int(reference_date.strftime("%Y%m%d")),
                time=reference_date.hour * 100,
                step=step,
                param=param,
                variable=variable,
                ensemble=self.ensemble,
                keys=keys,
                grib1_keys=self.grib1_keys,
                grib2_keys=self.grib2_keys,
                previous_step=previous_step,
                start_steps=start_steps,
            )

            for modifier in self.modifiers:
                values, template, keys = modifier(values, template, keys)

            if LOG.isEnabledFor(logging.DEBUG):
                LOG.info("Encoding GRIB %s\n%s", template, json.dumps(keys, indent=4))

            try:
                self.write_message(values, template=template, **keys)
            except Exception:
                LOG.error("Error writing field %s", name)
                LOG.error("Template: %s", template)
                LOG.error("Keys:\n%s", json.dumps(keys, indent=4, default=str))
                raise

    @abstractmethod
    def write_message(self, message: FloatArray, *args: Any, **kwargs: Any) -> None:
        """Write a message to the grib file.

        Parameters
        ----------
        message : FloatArray
            The message array.
        *args : Any
            Additional arguments.
        **kwargs : Any
            Additional keyword arguments.
        """
        pass

    def template(self, state: State, name: str) -> object:
        """Get the template for a variable.

        Parameters
        ----------
        state : State
            The state object.
        name : str
            The variable name.

        Returns
        -------
        object
            The template object.
        """

        if self.template_manager is None:
            self.template_manager = TemplateManager(self, self.templates)

        return self.template_manager.template(name, state)

    def template_lookup(self, name: str) -> dict:
        """Lookup the template for a variable.

        Parameters
        ----------
        name : str
            The variable name.

        Returns
        -------
        dict
            The template dictionary.
        """
        return self.encoding



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
import sys
from argparse import ArgumentParser
from argparse import Namespace
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional

from earthkit.data.utils.dates import to_datetime

from anemoi.inference.checkpoint import Checkpoint
from anemoi.inference.types import DataRequest
from anemoi.inference.types import Date

from ..config.run import RunConfiguration
from ..inputs.grib import GribInput
from ..inputs.mars import postproc
from ..runners import create_runner
from . import Command


def checkpoint_to_requests(
    checkpoint: Checkpoint,
    *,
    date: Date,
    target: Optional[str] = None,
    include_forcings: bool = True,
    retrieve_fields_type: Optional[str] = None,
    staging_dates: Optional[str] = None,
    use_grib_paramid: bool = False,
    extra: Optional[List[str]] = None,
    patch_request: Optional[Callable[[DataRequest], DataRequest]] = None,
    use_scda: bool = False,
) -> List[DataRequest]:
    """Convert a checkpoint to a list of data requests.

    Parameters
    ----------
    checkpoint : Checkpoint
        The checkpoint object containing the necessary data.
    date : Date
        The date for the data request.
    target : str, optional
        The target path for the data request.
    include_forcings : bool, optional
        Whether to include forcings in the data request.
    retrieve_fields_type : str, optional
        The type of fields to retrieve.
    staging_dates : str, optional
        Path to a file with staging dates.
    use_grib_paramid : bool, optional
        Whether to use GRIB parameter IDs.
    extra : list of str, optional
        Additional request values.
    patch_request : Callable[[DataRequest], DataRequest], optional
        Function to patch the data request.
    use_scda : bool, optional
        Whether to use SCDA stream for 6/18 input time.

    Returns
    -------
    List[DataRequest]
        A list of data requests.
    """
    # TODO: Move this to the runner

    variables = checkpoint.variables_from_input(include_forcings=include_forcings)
    area = checkpoint.area
    grid = checkpoint.grid

    if retrieve_fields_type is not None:
        selected = set()

        for name, kinds in checkpoint.variable_categories().items():
            if "computed" in kinds:
                continue
            for kind in kinds:
                if retrieve_fields_type.startswith(kind):  # PrepML adds an 's' to the type
                    selected.add(name)

        variables = sorted(selected)

    more = postproc(grid, area)

    for r in extra or []:
        k, v = r.split("=")
        more[k] = v

    if staging_dates:
        dates = set()
        with open(staging_dates) as f:
            for line in f:
                date = to_datetime(line.strip())
                dates.update([date + h for h in checkpoint.lagged])
        dates = sorted(dates)
    else:
        date = to_datetime(date)
        dates = [date + h for h in checkpoint.lagged]

    requests = []
    for r in checkpoint.mars_requests(
        dates=dates,
        variables=variables,
        use_grib_paramid=use_grib_paramid,
        patch_request=patch_request,
        always_split_time=use_scda,
    ):
        r = r.copy()
        if target is not None:
            r["target"] = target
        r.update(more)
        if use_scda:
            _patch_scda(r)
        requests.append(r)

    return requests


class RetrieveCmd(Command):
    """Used by prepml."""

    def add_arguments(self, command_parser: ArgumentParser) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : ArgumentParser
            The argument parser to which the arguments will be added.
        """
        command_parser.description = self.__doc__
        command_parser.add_argument(
            "config",
            type=str,
            help="Path to config file. Can be omitted to pass config with overrides and defaults.",
        )
        command_parser.add_argument("--defaults", action="append", help="Sources of default values.")
        command_parser.add_argument("--date", type=str, help="Date")
        command_parser.add_argument("--output", type=str, default=None, help="Output file")
        command_parser.add_argument("--staging-dates", type=str, help="Path to a file with staging dates")
        command_parser.add_argument("--extra", action="append", help="Additional request values. Can be repeated")
        command_parser.add_argument("--retrieve-fields-type", type=str, help="Type of fields to retrieve")
        command_parser.add_argument("--use-scda", action="store_true", help="Use scda stream for 6/18 input time")
        command_parser.add_argument("overrides", nargs="*", help="Overrides as key=value")

    def run(self, args: Namespace) -> None:
        """Run the retrieve command.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        if "=" in args.config:
            args.overrides.append(args.config)
            args.config = {}

        config: RunConfiguration = RunConfiguration.load(args.config, args.overrides, defaults=args.defaults)

        runner = create_runner(config)

        # so that the user does not need to pass --extra target=path when the input file is already in the config
        target = None
        input = runner.create_input()
        if isinstance(input, GribInput) and (path := getattr(input, "path", None)):
            target = path

        requests = checkpoint_to_requests(
            runner.checkpoint,
            date=args.date,
            target=target,
            include_forcings=True,
            extra=args.extra,
            retrieve_fields_type=args.retrieve_fields_type,
            staging_dates=args.staging_dates,
            use_grib_paramid=config.use_grib_paramid,
            patch_request=runner.patch_data_request,
            use_scda=args.use_scda,
        )

        if args.output and args.output != "-":
            f = open(args.output, "w")
        else:
            f = sys.stdout

        json.dump(requests, f, indent=4)


def _patch_scda(request: Dict[str, Any]) -> None:
    """Patch the SCDA stream in the request if necessary.
    ECMWF operational data has stream oper for 00 and 12 UTC and scda for 06 and 18 UTC.

    Parameters
    ----------
    request : dict
        The request dictionary to be patched.
    """

    if request.get("class", "od") != "od":
        return
    if request.get("type", "an") not in ("an", "fc"):
        return
    if request.get("stream", "oper") not in ("oper", "scda"):
        return

    request_time = int(request.get("time", 12))
    if request_time > 100:
        request_time = int(request_time / 100)

    if request_time in (6, 18):
        request["stream"] = "scda"


command = RetrieveCmd



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any
from typing import Dict

from anemoi.utils.logs import set_logging_name

from anemoi.inference.config import Configuration
from anemoi.inference.task import Task
from anemoi.inference.types import State

from ..transport import Transport
from . import transport_registry

LOG = logging.getLogger(__name__)


@transport_registry.register("mpi")
class MPITransport(Transport):
    """Transport implementation using MPI."""

    def __init__(self, couplings: Configuration, tasks: Dict[str, Task], *args: Any, **kwargs: Any) -> None:
        """Initialize the MPITransport.

        Parameters
        ----------
        couplings : Configuration
            The couplings for the transport.
        tasks : Dict[str, Any]
            The tasks to be executed.
        """
        from mpi4py import MPI

        super().__init__(couplings, tasks)
        self.comm: MPI.Comm = MPI.COMM_WORLD
        self.rank: int = self.comm.Get_rank()
        self.size: int = self.comm.Get_size()

        assert (
            len(tasks) == self.size
        ), f"Number of tasks ({len(tasks)}) must match number of MPI processes ({self.size})"

    def start(self) -> None:
        """Start the transport by initializing MPI tasks."""

        tasks = list(self.tasks.values())
        self.ranks = {task.name: i for i, task in enumerate(tasks)}

        # Pick only one task per rank
        task = tasks[self.rank]
        set_logging_name(task.name)

        task.run(self)

    def wait(self) -> None:
        """Wait for all MPI tasks to complete."""
        self.comm.barrier()

    def send(self, sender: Task, target: Task, state: State, tag: int) -> None:
        """Send a state from the sender to the target.

        Parameters
        ----------
        sender : Task
            The task sending the state.
        target : Task
            The task receiving the state.
        state : State
            The state to be sent.
        tag : int
            The tag associated with the state.
        """
        self.comm.send(state, dest=self.ranks[target.name], tag=tag)

    def receive(self, receiver: Task, source: Task, tag: int) -> State:
        """Receive a state from the source to the receiver.

        Parameters
        ----------
        receiver : Any
            The task receiving the state.
        source : Any
            The task sending the state.
        tag : int
            The tag associated with the state.

        Returns
        -------
        Any
            The received state.
        """
        return self.comm.recv(source=self.ranks[source.name], tag=tag)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os
from typing import Any
from typing import Dict
from typing import Optional

import earthkit.data as ekd

from . import IndexTemplateProvider
from . import template_provider_registry

LOG = logging.getLogger(__name__)


@template_provider_registry.register("samples")
class SamplesTemplates(IndexTemplateProvider):
    """Class to provide GRIB templates from sample files."""

    def load_template(self, grib: str, lookup: Dict[str, Any]) -> Optional[ekd.Field]:
        """Load a GRIB template based on the provided lookup dictionary.

        Parameters
        ----------
        grib : str
            The GRIB template string.
        lookup : Dict[str, Any]
            The lookup dictionary to format the GRIB template string.

        Returns
        -------
        Optional[ekd.Field]
            The loaded GRIB template field, or None if the template is not found.
        """
        template = grib.format(**lookup)
        if not os.path.exists(template):
            LOG.warning(f"Template not found: {template}")
            return None

        return ekd.from_source("file", template)[0]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from typing import Optional

from anemoi.inference.config import Configuration
from anemoi.inference.context import Context
from anemoi.inference.types import State

from ..output import ForwardOutput
from . import create_output
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("apply_mask")
class ApplyMaskOutput(ForwardOutput):
    """Apply mask output class.

    Parameters
    ----------
    context : dict
        The context dictionary.
    mask : str
        The mask identifier.
    output : dict
        The output configuration dictionary.
    output_frequency : int, optional
        The frequency of output, by default None.
    write_initial_state : bool, optional
        Whether to write the initial state, by default None.
    """

    def __init__(
        self,
        context: Context,
        *,
        mask: str,
        output: Configuration,
        output_frequency: Optional[int] = None,
        write_initial_state: Optional[bool] = None,
    ) -> None:
        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)
        self.mask = self.checkpoint.load_supporting_array(mask)
        self.output = create_output(context, output)

    def __repr__(self) -> str:
        """Return a string representation of the ApplyMaskOutput object."""
        return f"ApplyMaskOutput({self.mask}, {self.output})"

    def write_initial_step(self, state: State) -> None:
        """Write the initial step of the state.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        # Note: we foreward to 'state', so we write-up options again
        self.output.write_initial_state(self._apply_mask(state))

    def write_step(self, state: State) -> None:
        """Write a step of the state.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        # Note: we foreward to 'state', so we write-up options again
        self.output.write_state(self._apply_mask(state))

    def _apply_mask(self, state: State) -> State:
        """Apply the mask to the state.

        Parameters
        ----------
        state : State
            The state dictionary.

        Returns
        -------
        State
            The masked state dictionary.
        """
        state = state.copy()
        state["fields"] = state["fields"].copy()
        state["latitudes"] = state["latitudes"][self.mask]
        state["longitudes"] = state["longitudes"][self.mask]

        for field in state["fields"]:
            data = state["fields"][field]
            if data.ndim == 1:
                data = data[self.mask]
            else:
                data = data[..., self.mask]
            state["fields"][field] = data

        return state

    def close(self) -> None:
        """Close the output."""
        self.output.close()

    def print_summary(self, depth: int = 0) -> None:
        """Print the summary of the output.

        Parameters
        ----------
        depth : int, optional
            The depth of the summary, by default 0.
        """
        super().print_summary(depth)
        self.output.print_summary(depth + 1)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging
import os
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

import yaml
from pydantic import BaseModel
from pydantic import ConfigDict

LOG = logging.getLogger(__name__)

T = TypeVar("T", bound="Configuration")


class Configuration(BaseModel):
    """Configuration class."""

    model_config = ConfigDict(extra="forbid")

    @classmethod
    def load(
        cls: Type[T],
        path: Union[str, Dict[str, Any]],
        overrides: Union[List[str], List[dict], str, dict] = [],
        defaults: Optional[Union[str, List[str], dict]] = None,
    ) -> T:
        """Load the configuration.

        Parameters
        ----------
        path : Union[str, dict]
            Path to the configuration file or a dictionary containing the configuration.
        overrides : Union[List[str], List[dict], str, dict], optional
            List of overrides to apply to the configuration, by default [].
        defaults : Optional[Union[str, List[str], dict]], optional
            Default values to set in the configuration, by default None.

        Returns
        -------
        Configuration
            The loaded configuration.
        """

        config = {}

        # Set default values
        if defaults is not None:
            if not isinstance(defaults, list):
                defaults = [defaults]
            for d in defaults:
                if isinstance(d, str):
                    with open(d) as f:
                        d = yaml.safe_load(f)
                config.update(d)

        # Load the configuration
        if isinstance(path, dict):
            config.update(deepcopy(path))
        else:
            with open(path) as f:
                config.update(yaml.safe_load(f))

        # Apply overrides
        if not isinstance(overrides, list):
            overrides = [overrides]

        for override in overrides:
            if isinstance(override, dict):
                cls._merge_configs(config, override)
            else:
                path = config
                key, value = override.split("=")
                keys = key.split(".")
                for key in keys[:-1]:
                    if key.isdigit() and isinstance(path, list):
                        index = int(key)
                        if index < len(path):
                            LOG.debug(f"key {key} is used as list index in list{path}")
                            path = path[index]
                        elif index == len(path):
                            LOG.debug(f"key {key} is used to append to list {path}")
                            path.append({})
                            path = path[index]
                        else:
                            raise IndexError(f"Index {index} out of range for list {path} of length {len(path)}")
                    else:
                        path = path.setdefault(key, {})
                path[keys[-1]] = value

        # Validate the configuration
        config = cls(**config)

        # Set environment variables found in the configuration
        # as soon as possible
        for key, value in config.env.items():
            os.environ[key] = str(value)

        return config

    @classmethod
    def _merge_configs(cls, a: Dict[Any, Any], b: Dict[Any, Any]) -> None:
        """Merge two configurations.

        Parameters
        ----------
        a : Dict[Any, Any]
            The first configuration.
        b : Dict[Any, Any]
            The second configuration.
        """
        for key, value in b.items():
            if key in a and isinstance(a[key], dict) and isinstance(value, dict):
                cls._merge_configs(a[key], value)
            else:
                a[key] = value



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
from earthkit.data.utils.dates import to_datetime

from anemoi.inference.context import Context
from anemoi.inference.types import Date
from anemoi.inference.types import FloatArray
from anemoi.inference.types import State

from ..input import Input
from . import input_registry

LOG = logging.getLogger(__name__)


class DatasetInput(Input):
    """Handles `anemoi-datasets` dataset as input."""

    def __init__(self, context: Context, args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> None:
        """Initialize the DatasetInput.

        Parameters
        ----------
        context : Any
            The context in which the input is used.
        args : Tuple[Any, ...]
            Arguments for the dataset.
        kwargs : Dict[str, Any]
            Keyword arguments for the dataset.
        """
        super().__init__(context)

        grid_indices = kwargs.pop("grid_indices", None)

        self.args, self.kwargs = args, kwargs
        if context.verbosity > 0:
            LOG.info(
                "Opening dataset with\nargs=%s\nkwargs=%s", json.dumps(args, indent=4), json.dumps(kwargs, indent=4)
            )

        if grid_indices is None and "grid_indices" in context.checkpoint._supporting_arrays:
            grid_indices = context.checkpoint.load_supporting_array("grid_indices")
            if context.verbosity > 0:
                LOG.info(
                    "Loading supporting array `grid_indices` from checkpoint, \
                    the input grid will be reduced accordingly."
                )

        self.grid_indices = slice(None) if grid_indices is None else grid_indices

    @cached_property
    def ds(self) -> Any:
        """Return the dataset."""
        from anemoi.datasets import open_dataset

        return open_dataset(*self.args, **self.kwargs)

    @cached_property
    def latitudes(self) -> FloatArray:
        """Return the latitudes."""
        return self.ds.latitudes

    @cached_property
    def longitudes(self) -> FloatArray:
        """Return the longitudes."""
        return self.ds.longitudes

    def __repr__(self) -> str:
        """Return a string representation of the DatasetInput."""
        return f"DatasetInput({self.args}, {self.kwargs})"

    def create_input_state(self, *, date: Optional[Date] = None) -> State:
        """Create the input state for the given date.

        Parameters
        ----------
        date : Optional[Any]
            The date for which to create the input state.

        Returns
        -------
        Dict[str, Any]
            The created input state.
        """
        if date is None:
            raise ValueError("`date` must be provided")

        date = to_datetime(date)
        latitudes = self.ds.latitudes
        longitudes = self.ds.longitudes

        input_state = dict(
            date=date,
            latitudes=latitudes[self.grid_indices],
            longitudes=longitudes[self.grid_indices],
            fields=dict(),
        )

        fields = input_state["fields"]

        date = np.datetime64(date)
        data = self._load_dates([date + np.timedelta64(h) for h in self.checkpoint.lagged])

        if data.shape[2] != 1:
            raise ValueError(f"Ensemble data not supported, got {data.shape[2]} members")

        requested_variables = set(self.input_variables())
        for i, variable in enumerate(self.ds.variables):
            if variable not in requested_variables:
                continue
            # Squeeze the data to remove the ensemble dimension
            values = np.squeeze(data[:, i], axis=1)
            fields[variable] = values[:, self.grid_indices]

            if self.context.trace:
                self.context.trace.from_input(variable, self)

        return input_state

    def load_forcings_state(self, *, variables: List[str], dates: List[Date], current_state: State) -> State:
        """Load the forcings state for the given variables and dates.

        Parameters
        ----------
        variables : List[str]
            List of variables to load.
        dates : List[Any]
            List of dates for which to load the forcings.
        current_state : State
            The current state of the input.

        Returns
        -------
        State
            The loaded forcings state.
        """
        data = self._load_dates(dates)  # (date, variables, ensemble, values)

        requested_variables = np.array([self.ds.name_to_index[v] for v in variables])
        data = data[:, requested_variables]
        # Squeeze the data to remove the ensemble dimension
        data = np.squeeze(data, axis=2)
        # Reorder the dimensions to (variable, date, values)
        data = np.swapaxes(data, 0, 1)

        # apply reduction to `grid_indices`
        data = data[..., self.grid_indices]

        fields = {v: data[i] for i, v in enumerate(variables)}

        return dict(
            fields=fields,
            dates=dates,
            latitudes=self.latitudes,
            longitudes=self.longitudes,
        )

    def _load_dates(self, dates: List[Date]) -> Any:
        """Load the data for the given dates.

        Parameters
        ----------
        dates : List[Any]
            List of dates for which to load the data.

        Returns
        -------
        Any
            The loaded data.
        """
        # TODO: use the fact that the dates are sorted

        dataset_dates = self.ds.dates

        idx = []
        for d in dates:
            (i,) = np.where(dataset_dates == d)
            if len(i) == 0:
                raise ValueError(
                    f"Date {d} not found in the dataset, available dates: {dataset_dates[0]}...{dataset_dates[-1]} by {self.ds.frequency}"
                )
            assert len(i) == 1, f"Multiple dates found for {d}"
            idx.append(int(i[0]))

        if len(idx) == 1:
            s = slice(idx[0], idx[0] + 1)
        else:
            diff = idx[1] - idx[0]
            if not all(i == diff for i in np.diff(idx)):
                # TODO: remove that restriction
                raise ValueError("Dates do not have the same frequency")
            s = slice(idx[0], idx[-1] + 1, diff)

        return self.ds[s]


@input_registry.register("dataset")
class DatasetInputArgsKwargs(DatasetInput):
    """Handles `anemoi-datasets` dataset as input."""

    trace_name = "dataset/provided"

    def __init__(self, context: Context, /, *args: Any, use_original_paths: bool = True, **kwargs: Any) -> None:
        """Initialize the DatasetInputArgsKwargs.

        Parameters
        ----------
        context : Context
            The context in which the input is used.
        use_original_paths : bool
            Whether to use original paths.
        """
        if not args and not kwargs:
            args, kwargs = context.checkpoint.open_dataset_args_kwargs(use_original_paths=use_original_paths)

            # TODO: remove start/end from the arguments

            LOG.warning("No arguments provided to open_dataset, using the default arguments:")

            cmd = "open_dataset("
            for arg in args:
                cmd += f"{arg}, "
            for key, value in kwargs.items():
                cmd += f"{key}={value}, "

            LOG.warning("%s", cmd)
        super().__init__(context, args, kwargs)


class DataloaderInput(DatasetInput):
    """Handles `anemoi-datasets` dataset as input."""

    def __init__(self, context: Context, /, name: str, use_original_paths: bool = True, **kwargs: Any) -> None:
        """Initialize the DataloaderInput.

        Parameters
        ----------
        context : Any
            The context in which the input is used.
        name : str
            The name of the dataloader.
        use_original_paths : bool
            Whether to use original paths.
        """
        args, kwargs = context.checkpoint.open_dataset_args_kwargs(
            use_original_paths=use_original_paths,
            from_dataloader=name,
        )

        super().__init__(context, args, kwargs)


@input_registry.register("test")
class TestInput(DataloaderInput):
    """Handles `anemoi-datasets` dataset as input."""

    trace_name = "dataset/test"

    def __init__(self, context: Context, /, use_original_paths: bool = True, **kwargs: Any) -> None:
        """Initialize the TestInput.

        Parameters
        ----------
        context : Any
            The context in which the input is used.
        use_original_paths : bool
            Whether to use original paths.
        """
        super().__init__(
            context,
            name="test",
            use_original_paths=use_original_paths,
            **kwargs,
        )


@input_registry.register("validation")
class ValidationInput(DataloaderInput):
    """Handles `anemoi-datasets` dataset as input."""

    trace_name = "dataset/validation"

    def __init__(self, context: Context, /, use_original_paths: bool = True, **kwargs: Any) -> None:
        super().__init__(
            context,
            name="validation",
            use_original_paths=use_original_paths,
            **kwargs,
        )


@input_registry.register("training")
class TrainingInput(DataloaderInput):
    """Handles `anemoi-datasets` dataset as input."""

    trace_name = "dataset/training"

    def __init__(self, context: Context, /, use_original_paths: bool = True, **kwargs: Any) -> None:
        super().__init__(
            context,
            name="training",
            use_original_paths=use_original_paths,
            **kwargs,
        )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any
from typing import List
from typing import Optional

import earthkit.data as ekd

from anemoi.inference.context import Context
from anemoi.inference.types import Date
from anemoi.inference.types import State

from . import input_registry
from .grib import GribInput

LOG = logging.getLogger(__name__)


@input_registry.register("icon_grib_file")
class IconInput(GribInput):
    """Handles grib files from ICON."""

    # TODO: this code will become a plugin in the future.

    trace_name = "icon file"

    def __init__(
        self,
        context: Context,
        path: str,
        grid: str,
        refinement_level_c: int,
        namer: Optional[Any] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the IconInput.

        Parameters
        ----------
        context : Any
            The context in which the input is used.
        path : str
            The path to the ICON grib file.
        grid : str
            The grid type.
        refinement_level_c : int
            The refinement level.
        namer : Optional[Any]
            Optional namer for the input.
        **kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context, namer=namer, **kwargs)
        self.path = path
        self.grid = grid
        self.refinement_level_c = refinement_level_c

    def create_input_state(self, *, date: Optional[Date]) -> State:
        """Creates the input state for the given date.

        Parameters
        ----------
        date : Optional[Date]
            The date for which to create the input state.

        Returns
        -------
        State
            The created input state.
        """
        from anemoi.transform.grids.icon import icon_grid

        latitudes, longitudes = icon_grid(self.grid, self.refinement_level_c)

        return self._create_input_state(
            ekd.from_source("file", self.path),
            variables=None,
            date=date,
            latitudes=latitudes,
            longitudes=longitudes,
        )

    def load_forcings_state(self, *, variables: List[str], dates: List[Date], current_state: State) -> State:
        """Loads the forcings state for the given variables and dates.

        Parameters
        ----------
        variables : List[str]
            List of variable names.
        dates : List[Date]
            List of dates for which to load the forcings state.
        current_state : State
            The current state.

        Returns
        -------
        Any
            The loaded forcings state.
        """
        return self._load_forcings_state(
            ekd.from_source("file", self.path),
            variables=variables,
            dates=dates,
            current_state=current_state,
        )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any
from typing import List

from anemoi.inference.types import IntArray
from anemoi.inference.types import State

from ..forcings import ComputedForcings
from ..forcings import Forcings
from ..runner import Runner
from . import runner_registry

LOG = logging.getLogger(__name__)

# This is because forcings are assumed to already be in the
# state dictionary, so we don't need to load them from anywhere.


class NoForcings(Forcings):
    """No forcings."""

    def __init__(self, context: Any, variables: List[str], mask: IntArray) -> None:
        """Initialize the NoForcings.

        Parameters
        ----------
        context : object
            The context for the forcings.
        variables : list
            The variables for the forcings.
        mask : IntArray
            The mask for the forcings.
        """
        super().__init__(context)
        self.variables = variables
        self.mask = mask
        self.kinds = dict(unknown=True)

    def load_forcings_state(self, state: State, date: str) -> None:
        """Load forcings state.

        Parameters
        ----------
        state : State
            The state to load the forcings into.
        date : str
            The date for which to load the forcings.
        """
        pass


@runner_registry.register("simple")
class SimpleRunner(Runner):
    """Use that runner when using the low level API."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the SimpleRunner.

        Parameters
        ----------
        *args : tuple
            Positional arguments.
        **kwargs : dict
            Keyword arguments.
        """
        super().__init__(*args, **kwargs)

    def create_constant_computed_forcings(self, variables: List[str], mask: IntArray) -> List[Forcings]:
        """Create constant computed forcings.

        Parameters
        ----------
        variables : List[str]
            The variables for the forcings.
        mask : IntArray
            The mask for the forcings.

        Returns
        -------
        List[Forcings]
            The created constant computed forcings.
        """
        result = ComputedForcings(self, variables, mask)
        LOG.info("Constant computed forcing: %s", result)
        return [result]

    def create_dynamic_computed_forcings(self, variables: List[str], mask: IntArray) -> List[Forcings]:
        """Create dynamic computed forcings.

        Parameters
        ----------
        variables : List[str]
            The variables for the forcings.
        mask : IntArray
            The mask for the forcings.

        Returns
        -------
        List[Forcings]
            The created dynamic computed forcings.
        """
        result = ComputedForcings(self, variables, mask)
        LOG.info("Dynamic computed forcing: %s", result)
        return [result]

    def create_constant_coupled_forcings(self, variables: List[str], mask: IntArray) -> List[Forcings]:
        """Create constant coupled forcings.

        Parameters
        ----------
        variables : List[str]
            The variables for the forcings.
        mask : IntArray
            The mask for the forcings.

        Returns
        -------
        List
            The created constant coupled forcings.
        """
        # This runner does not support coupled forcings
        # there are supposed to be already in the state dictionary
        # or managed by the user.
        LOG.warning("Coupled forcings are not supported by this runner: %s", variables)
        return []

    def create_dynamic_coupled_forcings(self, variables: List[str], mask: IntArray) -> List[Forcings]:
        """Create dynamic coupled forcings.

        Parameters
        ----------
        variables : List[str]
            The variables for the forcings.
        mask : IntArray
            The mask for the forcings.

        Returns
        -------
        List
            The created dynamic coupled forcings.
        """
        # This runner does not support coupled forcings
        # there are supposed to be already in the state dictionary
        # or managed by the user.
        LOG.warning("Coupled forcings are not supported by this runner: %s", variables)
        return []



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import warnings
from typing import Any
from typing import Dict
from typing import List

from anemoi.utils.config import DotDict
from anemoi.utils.dates import frequency_to_timedelta as to_timedelta
from pydantic import BaseModel

from anemoi.inference.config import Configuration
from anemoi.inference.input import Input
from anemoi.inference.output import Output
from anemoi.inference.processor import Processor
from anemoi.inference.types import IntArray
from anemoi.inference.types import State

from ..forcings import BoundaryForcings
from ..forcings import ComputedForcings
from ..forcings import CoupledForcings
from ..forcings import Forcings
from ..inputs import create_input
from ..outputs import create_output
from ..post_processors import create_post_processor
from ..pre_processors import create_pre_processor
from ..runner import Runner
from . import runner_registry

LOG = logging.getLogger(__name__)


@runner_registry.register("default")
class DefaultRunner(Runner):
    """Default runner class for inference.

    This class provides the default implementation for running inference.
    """

    def __init__(self, config: Configuration) -> None:
        """Initialize the DefaultRunner.

        Parameters
        ----------
        config : Configuration
            Configuration for the runner.
        """

        if isinstance(config, dict):
            # So we get the dot notation
            config = DotDict(config)

        # Remove that when the Pydantic model is ready
        if isinstance(config, BaseModel):
            config = DotDict(config.model_dump())

        self.config = config

        super().__init__(
            config.checkpoint,
            device=config.device,
            precision=config.precision,
            allow_nans=config.allow_nans,
            verbosity=config.verbosity,
            report_error=config.report_error,
            use_grib_paramid=config.use_grib_paramid,
            patch_metadata=config.patch_metadata,
            development_hacks=config.development_hacks,
            output_frequency=config.output_frequency,
            write_initial_state=config.write_initial_state,
            trace_path=config.trace_path,
            use_profiler=config.use_profiler,
        )

    def execute(self) -> None:
        """Execute the runner."""

        if self.config.description is not None:
            LOG.info("%s", self.config.description)

        lead_time = to_timedelta(self.config.lead_time)

        # This may be used by Output objects to compute the step
        self.lead_time = lead_time
        self.time_step = self.checkpoint.timestep

        input = self.create_input()
        output = self.create_output()

        # pre_processors = self.pre_processors
        post_processors = self.post_processors

        input_state = input.create_input_state(date=self.config.date)

        # This hook is needed for the coupled runner
        self.input_state_hook(input_state)

        state = Output.reduce(input_state)
        for processor in post_processors:
            state = processor.process(state)

        output.open(state)
        output.write_initial_state(state)

        for state in self.run(input_state=input_state, lead_time=lead_time):
            for processor in post_processors:
                LOG.info("Post processor: %s", processor)
                state = processor.process(state)
            output.write_state(state)

        output.close()

    def input_state_hook(self, input_state: State) -> None:
        """Hook used by coupled runners to send the input state."""
        pass

    def create_input(self) -> Input:
        """Create the input.

        Returns
        -------
        Input
            The created input.
        """
        input = create_input(self, self.config.input)
        LOG.info("Input: %s", input)
        return input

    def create_output(self) -> Output:
        """Create the output.

        Returns
        -------
        Output
            The created output.
        """
        output = create_output(self, self.config.output)
        LOG.info("Output:")
        output.print_summary()
        return output

    def create_constant_computed_forcings(self, variables: List[str], mask: IntArray) -> List[Forcings]:
        """Create constant computed forcings.

        Parameters
        ----------
        variables : List[str]
            The variables for the forcings.
        mask : IntArray
            The mask for the forcings.

        Returns
        -------
        List[Forcings]
            The created constant computed forcings.
        """
        result = ComputedForcings(self, variables, mask)
        LOG.info("Constant computed forcing: %s", result)
        return [result]

    def create_dynamic_computed_forcings(self, variables: List[str], mask: IntArray) -> List[Forcings]:
        """Create dynamic computed forcings.

        Parameters
        ----------
        variables : List[str]
            The variables for the forcings.
        mask : IntArray
            The mask for the forcings.

        Returns
        -------
        List[Forcings]
            The created dynamic computed forcings.
        """
        result = ComputedForcings(self, variables, mask)
        LOG.info("Dynamic computed forcing: %s", result)
        return [result]

    def _input_forcings(self, name: str) -> Dict[str, Any]:
        """Get the input forcings configuration.

        Parameters
        ----------
        name : str
            The name of the forcings configuration.

        Returns
        -------
        dict
            The input forcings configuration.
        """
        if self.config.forcings is None:
            # Use the same as the input
            return self.config.input

        if name in self.config.forcings:
            return self.config.forcings[name]

        if "input" in self.config.forcings:
            return self.config.forcings.input

        return self.config.forcings

    def create_constant_coupled_forcings(self, variables: List[str], mask: IntArray) -> List[Forcings]:
        """Create constant coupled forcings.

        Parameters
        ----------
        variables : list
            The variables for the forcings.
        mask : IntArray
            The mask for the forcings.

        Returns
        -------
        List[Forcings]
            The created constant coupled forcings.
        """
        input = create_input(self, self._input_forcings("constant"))
        result = CoupledForcings(self, input, variables, mask)
        LOG.info("Constant coupled forcing: %s", result)
        return [result]

    def create_dynamic_coupled_forcings(self, variables: List[str], mask: IntArray) -> List[Forcings]:
        """Create dynamic coupled forcings.

        Parameters
        ----------
        variables : list
            The variables for the forcings.
        mask : IntArray
            The mask for the forcings.

        Returns
        -------
        List[Forcings]
            The created dynamic coupled forcings.
        """
        input = create_input(self, self._input_forcings("dynamic"))
        result = CoupledForcings(self, input, variables, mask)
        LOG.info("Dynamic coupled forcing: %s", result)
        return [result]

    def create_boundary_forcings(self, variables: List[str], mask: IntArray) -> List[Forcings]:
        """Create boundary forcings.

        Parameters
        ----------
        variables : list
            The variables for the forcings.
        mask : IntArray
            The mask for the forcings.

        Returns
        -------
        List[Forcings]
            The created boundary forcings.
        """
        input = create_input(self, self._input_forcings("boundary"))
        result = BoundaryForcings(self, input, variables, mask)
        LOG.info("Boundary forcing: %s", result)
        return [result]

    def create_pre_processors(self) -> List[Processor]:
        """Create pre-processors.

        Returns
        -------
        List[Processor]
            The created pre-processors.
        """
        if self.config.post_processors is None:
            self.config.post_processors = ["accumulate_from_start_of_forecast"]
            warnings.warn(
                """
                No post_processors defined. Accumulations will be accumulated from the beginning of the forecast.

                🚧🚧🚧 In a future release, the default will be to NOT accumulate from the beginning of the forecast. 🚧🚧🚧
                Update your config if you wish to keep accumulating from the beginning.
                https://github.com/ecmwf/anemoi-inference/issues/131
                """,
            )

        if "accumulate_from_start_of_forecast" not in self.config.post_processors:
            warnings.warn(
                """
                post_processors are defined but `accumulate_from_start_of_forecast` is not set."
                🚧 Accumulations will NOT be accumulated from the beginning of the forecast. 🚧
                """
            )

        result = []
        for processor in self.config.pre_processors:
            result.append(create_pre_processor(self, processor))

        LOG.info("Pre processors: %s", result)
        return result

    def create_post_processors(self) -> List[Processor]:
        """Create post-processors.

        Returns
        -------
        List[Processor]
            The created post-processors.
        """
        result = []
        for processor in self.config.post_processors:
            result.append(create_post_processor(self, processor))

        LOG.info("Post processors: %s", result)
        return result



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""Dummy input used for testing.

It will generate fields with constant values for each variable and date.
These values are then tested in the mock model.
"""

import datetime
import logging
from typing import Any
from typing import List
from typing import Optional

import earthkit.data as ekd
import numpy as np

from anemoi.inference.context import Context
from anemoi.inference.testing import float_hash
from anemoi.inference.types import Date
from anemoi.inference.types import State

from . import input_registry
from .ekd import EkdInput

LOG = logging.getLogger(__name__)
SKIP_KEYS = ["date", "time", "step"]


@input_registry.register("dummy")
class DummyInput(EkdInput):
    """Dummy input used for testing."""

    trace_name = "dummy"

    def __init__(self, context: Context, *, namer: Optional[Any] = None, **kwargs: Any) -> None:
        """Initialize the DummyInput.

        Parameters
        ----------
        context : Context
            The context in which the input is used.
        namer : Optional[Any]
            Optional namer for the input.
        **kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context, namer=namer, **kwargs)

    def create_input_state(self, *, date: Optional[Date]) -> State:
        """Create the input state for the given date.

        Parameters
        ----------
        date : Optional[Date]
            The date for which to create the input state.

        Returns
        -------
        State
            The created input state.
        """
        if date is None:
            date = datetime.datetime(2000, 1, 1)

        dates = [date + h for h in self.checkpoint.lagged]
        return self._create_input_state(self._fields(dates), variables=None, date=date)

    def load_forcings_state(self, *, variables: List[str], dates: List[Date], current_state: State) -> State:
        """Load the forcings state for the given variables and dates.

        Parameters
        ----------
        variables : List[str]
            List of variables to load.
        dates : List[Date]
            List of dates for which to load the forcings.
        current_state : State
            The current state of the input.

        Returns
        -------
        Any
            The loaded forcings state.
        """
        return self._load_forcings_state(
            self._fields(dates),
            variables=variables,
            dates=dates,
            current_state=current_state,
        )

    def _fields(self, dates: List[Date], variables: Optional[List[str]] = None) -> ekd.FieldList:
        """Generate fields for the given dates and variables.

        Parameters
        ----------
        dates : List[Date]
            List of dates for which to generate fields, by default None.
        variables : Optional[List[str]], optional
            List of variables for which to generate fields, by default None.

        Returns
        -------
        ekd.FieldList
            The generated fields.
        """

        if variables is None:
            variables = self.checkpoint.variables_from_input(include_forcings=True)

        LOG.info("Generating fields for %s", variables)

        typed_variables = self.checkpoint.typed_variables

        result = []
        for variable in variables:
            is_constant_in_time = typed_variables[variable].is_constant_in_time

            keys = {k: v for k, v in typed_variables[variable].grib_keys.items() if k not in SKIP_KEYS}

            for date in dates:
                x = float_hash(variable, dates[0] if is_constant_in_time else date)

                handle = dict(
                    values=np.ones(self.checkpoint.number_of_grid_points, dtype=np.float32) * x,
                    latitudes=np.zeros(self.checkpoint.number_of_grid_points, dtype=np.float32),
                    longitudes=np.zeros(self.checkpoint.number_of_grid_points, dtype=np.float32),
                    date=date.strftime("%Y%m%d"),
                    time=date.strftime("%H%M"),
                    name=variable,
                    **keys,
                )
                result.append(handle)

        return ekd.from_source("list-of-dicts", result)

    def template_lookup(self, name: str) -> dict:
        """Lookup a template by name.

        Parameters
        ----------
        name : str
            The name of the template to lookup.

        Returns
        -------
        dict
            The template dictionary.
        """
        # Unused, but required by the TemplateManager
        return {}



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging
from argparse import ArgumentParser
from argparse import Namespace

from ..config.couple import CoupleConfiguration
from ..tasks import create_task
from ..transports import create_transport
from . import Command

LOG = logging.getLogger(__name__)

COPY_ATTRIBUTES = (
    "date",
    "lead_time",
    "verbosity",
    "report_error",
)


class CoupleCmd(Command):
    """Couple tasks based on a configuration file."""

    def add_arguments(self, command_parser: ArgumentParser) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : ArgumentParser
            The argument parser to which the arguments will be added.
        """
        command_parser.add_argument("--defaults", action="append", help="Sources of default values.")
        command_parser.add_argument(
            "config",
            help="Path to config file. Can be omitted to pass config with overrides and defaults.",
        )
        command_parser.add_argument("overrides", nargs="*", help="Overrides as key=value")

    def run(self, args: Namespace) -> None:
        """Run the couple command.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        if "=" in args.config:
            args.overrides.append(args.config)
            args.config = {}

        config = CoupleConfiguration.load(
            args.config,
            args.overrides,
            defaults=args.defaults,
        )

        if config.description is not None:
            LOG.info("%s", config.description)

        global_config = {}
        for copy in COPY_ATTRIBUTES:
            value = getattr(config, copy, None)
            if value is not None:
                LOG.info("Copy setting to all tasks: %s=%s", copy, value)
                global_config[copy] = value

        tasks = {name: create_task(name, action, global_config=global_config) for name, action in config.tasks.items()}
        for task in tasks.values():
            LOG.info("Task: %s", task)

        transport = create_transport(config.transport, config.couplings, tasks)
        LOG.info("Transport: %s", transport)

        transport.start()
        transport.wait()
        LOG.info("Run complete")


command = CoupleCmd



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import functools
import os
from typing import Any
from typing import Callable


def fake_checkpoints(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to mock checkpoints for testing.

    Parameters
    ----------
    func : Callable
        The function to be decorated.

    Returns
    -------
    Callable
        The decorated function.
    """

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        from unittest.mock import MagicMock
        from unittest.mock import patch

        from .mock_checkpoint import MockRunConfiguration
        from .mock_checkpoint import mock_load_metadata
        from .mock_checkpoint import mock_torch_load

        with (
            patch("anemoi.inference.checkpoint.load_metadata", mock_load_metadata),
            patch("anemoi.inference.provenance.validate_environment", MagicMock()),
            patch("torch.load", mock_torch_load),
            patch("anemoi.inference.metadata.USE_LEGACY", True),
            patch("anemoi.inference.tasks.runner.RunConfiguration", MockRunConfiguration),
        ):
            return func(*args, **kwargs)

    return wrapper


def float_hash(s: str, date: datetime.datetime, accuracy: int = 1_000_000) -> float:
    """Hash a string and date to a float.

    Parameters
    ----------
    s : str
        The string to be hashed.
    date : datetime.datetime
        The date to be hashed.
    accuracy : int, optional
        The accuracy of the hash, by default 1_000_000

    Returns
    -------
    float
        The resulting hash as a float.
    """
    h = s + date.isoformat()
    return float(int.from_bytes(h.encode(), "little") % accuracy) / accuracy


def files_for_tests(name: str) -> str:
    """Get the path to the testing files.

    Parameters
    ----------
    name : str
        The name of the test file.

    Returns
    -------
    str
        The path to the test file.
    """
    # Running in GitHub Actions, the package was installed

    if "GITHUB_WORKSPACE" in os.environ:
        return os.path.join(os.environ["GITHUB_WORKSPACE"], "tests", name)

    # We assume that the test data is in the same directory as the test
    # and this is a development environment

    import anemoi.inference

    bits = os.path.normpath(anemoi.inference.__file__).split(os.path.sep)
    while len(bits) > 3 and (bits[-3], bits[-2], bits[-1]) != ("src", "anemoi", "inference"):
        bits.pop()

    for i in range(3):
        bits.pop()

    bits.append("tests")
    bits.append(name)
    return os.path.sep.join(bits)


class TestingContext:
    """A context for testing plugins."""

    pass



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import logging
import sys
from typing import Any
from typing import Dict
from typing import Union

import numpy as np

from anemoi.inference.types import BoolArray
from anemoi.inference.types import FloatArray

LOG = logging.getLogger(__name__)


class RolloutSource:
    """Represents a source of data that is a rollout."""

    trace_name = "rollout"


class UnknownSource:
    """Represents a source of data that is unknown."""

    trace_name = "?"


class UnchangedSource:
    """Represents a source of data that is unchanged."""

    trace_name = "unchanged"


class InputSource:
    """Represents a source of data that is an input."""

    def __init__(self, input: Any) -> None:
        self.input = input
        self.trace_name = input.trace_name


class Trace:
    """Implementation of a trace."""

    def __init__(self, path: Union[bool, str]) -> None:
        self.path = path
        self.file = sys.stdout if path is True else open(path, "w")

        self.sources: Dict[str, Any] = {}
        self.extra: Dict[str, Any] = {}

        print("-+" * 80, file=self.file)
        print("Trace:", datetime.datetime.now(), file=self.file)
        print("-+" * 80, file=self.file)
        print(file=self.file, flush=True)

    def write_input_tensor(
        self,
        date: datetime.datetime,
        fcstep: int,
        input_tensor: FloatArray,
        variable_to_input_tensor_index: Dict[str, int],
        timestep: datetime.timedelta,
    ) -> None:
        """Write the input tensor details to the trace file.

        Parameters
        ----------
        date : datetime.datetime
            The date associated with the input tensor.
        fcstep : int
            The step number.
        input_tensor : FloatArray
            The input tensor.
        variable_to_input_tensor_index : Dict[str, int]
            Mapping from variable names to input tensor indices.
        timestep : datetime.timedelta
            The timestep.
        """
        print(f"Input tensor to forecast date {date-timestep} => {date}", input_tensor.shape, file=self.file)
        names = {v: k for k, v in variable_to_input_tensor_index.items()}
        assert len(input_tensor.shape) == 4
        assert input_tensor.shape[0] == 1
        lines = []

        for s in range(input_tensor.shape[1]):
            for i in range(input_tensor.shape[-1]):
                values = input_tensor[0, s, :, i]
                if s == 0:
                    lines.append(
                        [names[i], np.nanmin(values), np.nanmax(values), np.nanmean(values), np.nanstd(values)]
                    )
                else:
                    constant = np.all(values == input_tensor[0, 0, :, i])
                    constant = "C" if constant else ""
                    lines[i].extend(
                        [np.nanmin(values), np.nanmax(values), np.nanmean(values), np.nanstd(values), constant]
                    )

        unknown = UnknownSource()
        for line in lines:
            line.append(self.sources.get(line[0], unknown).trace_name)

        lines.insert(0, ["--------", "---", "---", "----", "---", "---", "---", "----", "---", "-", "---"])
        lines.insert(0, ["Variable", "Min", "Max", "Mean", "Std", "Min", "Max", "Mean", "Std", "C", "Source"])
        self.table(lines)

        print(file=self.file, flush=True)

    def write_output_tensor(
        self,
        date: datetime.datetime,
        fcstep: int,
        output_tensor: FloatArray,
        output_tensor_index_to_variable: Dict[int, str],
        timestep: datetime.timedelta,
    ) -> None:
        """Write the output tensor details to the trace file.

        Parameters
        ----------
        date : datetime.datetime
            The date associated with the output tensor.
        fcstep : int
            The step number.
        output_tensor : FloatArray
            The output tensor.
        output_tensor_index_to_variable : Dict[int, str]
            Mapping from output tensor indices to variable names.
        timestep : datetime.timedelta
            The timestep.
        """
        print(f"Output tensor for {date}:", output_tensor.shape, file=self.file)
        assert len(output_tensor.shape) == 2
        names = output_tensor_index_to_variable
        lines = []
        for i in range(output_tensor.shape[-1]):
            values = output_tensor[:, i]
            lines.append([names[i], np.nanmin(values), np.nanmax(values), np.nanmean(values), np.nanstd(values)])

        lines.insert(0, ["--------", "---", "---", "----", "---"])
        lines.insert(0, ["Variable", "Min", "Max", "Mean", "Std"])
        self.table(lines)
        print(file=self.file, flush=True)

    def table(self, lines: list[list[Any]]) -> None:
        """Print a formatted table to the trace file.

        Parameters
        ----------
        lines : list[list[Any]]
            The table data to print.
        """
        print(file=self.file)

        def _(x: Any) -> str:
            if isinstance(x, float):
                return f"{x:g}"
            return str(x)

        lines = [[_(x) for x in line] for line in lines]

        lengths = [max(len(str(x)) for x in col) for col in zip(*lines)]
        for line in lines:
            print(" ".join(f"{x:{lengths[i]}}" for i, x in enumerate(line)), file=self.file)

    def from_source(self, name: str, source: Any, extra: Any = None) -> None:
        """Add a source to the trace.

        Parameters
        ----------
        name : str
            The name of the source.
        source : Any
            The source object.
        extra : Any, optional
            Additional information about the source.
        """
        if name in self.sources:
            old = self.sources[name]
            LOG.warning(
                f"Warning: source {name} already defined as '{old.trace_name}', changing to '{source.trace_name}'"
            )

        # assert name not in self.sources, name
        self.sources[name] = source
        self.extra[name] = extra

    def from_rollout(self, name: str) -> None:
        """Add a rollout source to the trace.

        Parameters
        ----------
        name : str
            The name of the rollout source.
        """
        self.from_source(name, RolloutSource())

    def from_input(self, name: str, input: Any) -> None:
        """Add an input source to the trace.

        Parameters
        ----------
        name : str
            The name of the input source.
        input : Any
            The input object.
        """
        self.from_source(name, InputSource(input))

    def reset_sources(self, reset: BoolArray, variable_to_input_tensor_index: Dict[str, int]) -> None:
        """Reset the sources in the trace.

        Parameters
        ----------
        reset : list[bool]
            List indicating which sources to reset.
        variable_to_input_tensor_index : Dict[str, int]
            Mapping from variable names to input tensor indices.
        """
        self.sources = {}
        self.extra = {}
        names = {v: k for k, v in variable_to_input_tensor_index.items()}
        for i, on in enumerate(reset):
            if on:
                self.sources[names[i]] = UnchangedSource()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any
from typing import Optional

from ..context import Context
from ..input import Input
from ..runner import Runner
from . import runner_registry

LOG = logging.getLogger(__name__)


class CutoutInput(Input):
    """An Input object that combines two inputs."""

    def __init__(self, context: Context, lam: Input, globe: Input) -> None:
        """Initialize CutoutInput.

        Parameters
        ----------
        context : Context
            The context for the input.
        lam : Input
            The LAM input.
        globe : Input
            The globe input.
        """
        super().__init__(context)
        self.lam = lam
        self.globe = globe

    def create_input_state(self, *, date: Optional[str] = None) -> None:
        """Create the input state.

        Parameters
        ----------
        date : Optional[str], optional
            The date for the input state, by default None
        """
        state1 = self.lam.create_input_state(date=date)
        state2 = self.globe.create_input_state(date=date)

        assert False, (state1, state2)


@runner_registry.register("cutout")
class CutoutRunner(Runner):
    """A Runner that for LAMs."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize CutoutRunner."""
        super().__init__(*args, **kwargs)

        sources = self.checkpoint.sources

        if len(sources) != 2:
            raise ValueError(f"CutoutRunner expects two source, found {len(sources)}.")

        self.lam, self.globe = sources



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
import logging
import os
import subprocess
from argparse import ArgumentParser
from argparse import Namespace
from tempfile import TemporaryDirectory
from typing import Any
from typing import Dict

import yaml

from . import Command

LOG = logging.getLogger(__name__)


class Metadata(Command):
    """Edit, remove, dump or load metadata from a checkpoint file."""

    def add_arguments(self, command_parser: ArgumentParser) -> None:
        """Add command line arguments to the parser.

        Parameters
        ----------
        command_parser : ArgumentParser
            The argument parser to which the arguments will be added.
        """
        from anemoi.utils.checkpoints import DEFAULT_NAME

        command_parser.add_argument("path", help="Path to the checkpoint.")

        group = command_parser.add_mutually_exclusive_group(required=True)

        group.add_argument(
            "--dump",
            action="store_true",
            help=(
                "Extract the metadata from the checkpoint and print it to the standard output"
                " or the file specified by ``--output``, in JSON or YAML format."
            ),
        )
        group.add_argument(
            "--load",
            action="store_true",
            help=(
                "Set the metadata in the checkpoint from the content"
                " of a file specified by the ``--input`` argument."
            ),
        )

        group.add_argument(
            "--edit",
            action="store_true",
            help=(
                "Edit the metadata in place, using the specified editor."
                " See the ``--editor`` argument for more information."
            ),
        )

        group.add_argument(
            "--view",
            action="store_true",
            help=(
                "View the metadata in place, using the specified pager."
                " See the ``--pager`` argument for more information."
            ),
        )

        group.add_argument(
            "--remove",
            action="store_true",
            help="Remove the metadata from the checkpoint.",
        )

        group.add_argument(
            "--supporting-arrays",
            action="store_true",
            help="Print the supporting arrays.",
        )

        group.add_argument(
            "--get",
            help="Navigate the metadata via dot-separated path.",
        )

        group.add_argument(
            "--pytest",
            action="store_true",
            help=("Extract the metadata from the checkpoint so it can be added to the test suite."),
        )

        command_parser.add_argument(
            "--name",
            default=DEFAULT_NAME,
            help="Name of metadata record to be used with the actions above.",
        )

        command_parser.add_argument(
            "--input",
            help="The output file name to be used by the ``--load`` option.",
        )

        command_parser.add_argument(
            "--output",
            help="The output file name to be used by the ``--dump`` option.",
        )

        command_parser.add_argument(
            "--editor",
            help="Editor to use for the ``--edit`` option. Default to ``$EDITOR`` if defined, else ``vi``.",
            default=os.environ.get("EDITOR", "vi"),
        )

        command_parser.add_argument(
            "--pager",
            help="Editor to use for the ``--view`` option. Default to ``$PAGER`` if defined, else ``less``.",
            default=os.environ.get("PAGER", "less"),
        )

        command_parser.add_argument(
            "--json",
            action="store_true",
            help="Use the JSON format with ``--dump``, ``--view`` and ``--edit``.",
        )

        command_parser.add_argument(
            "--yaml",
            action="store_true",
            help="Use the YAML format with ``--dump``, ``--view`` and ``--edit``.",
        )

    def run(self, args: Namespace) -> None:
        """Execute the command based on the provided arguments.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        if args.edit:
            return self.edit(args)

        if args.view:
            return self.view(args)

        if args.get:
            return self.get(args)

        if args.remove:
            return self.remove(args)

        if args.dump or args.pytest:
            return self.dump(args)

        if args.load:
            return self.load(args)

        if args.supporting_arrays:
            return self.supporting_arrays(args)

    def edit(self, args: Namespace) -> None:
        """Edit the metadata in place using the specified editor.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        return self._edit(args, view=False, cmd=args.editor)

    def view(self, args: Namespace) -> None:
        """View the metadata in place using the specified pager.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        return self._edit(args, view=True, cmd=args.pager)

    def _edit(self, args: Namespace, view: bool, cmd: str) -> None:
        """Internal method to edit or view the metadata.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        view : bool
            If True, view the metadata; otherwise, edit it.
        cmd : str
            The command to use for editing or viewing.
        """
        from anemoi.utils.checkpoints import load_metadata
        from anemoi.utils.checkpoints import replace_metadata

        kwargs: Dict[str, Any] = {}

        if args.json:
            ext = "json"
            dump = json.dump
            load = json.load
            if args.test:
                kwargs = {"sort_keys": True}
            else:
                kwargs = {"indent": 4, "sort_keys": True}
        else:
            ext = "yaml"
            dump = yaml.dump
            load = yaml.safe_load
            kwargs = {"default_flow_style": False}

        with TemporaryDirectory() as temp_dir:

            path = os.path.join(temp_dir, f"checkpoint.{ext}")
            metadata = load_metadata(args.path)

            with open(path, "w") as f:
                dump(metadata, f, **kwargs)

            subprocess.check_call([cmd, path])

            if not view:
                with open(path) as f:
                    edited = load(f)

                if edited != metadata:
                    replace_metadata(args.path, edited)
                else:
                    LOG.info("No changes made.")

    def remove(self, args: Namespace) -> None:
        """Remove the metadata from the checkpoint.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        from anemoi.utils.checkpoints import remove_metadata

        remove_metadata(args.path, args.name)

    def dump(self, args: Namespace) -> None:
        """Dump the metadata from the checkpoint to a file or standard output.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        from anemoi.utils.checkpoints import load_metadata

        if args.output:
            file = open(args.output, "w")
        else:
            file = None

        metadata = load_metadata(args.path)
        if args.pytest:
            from anemoi.inference.testing.mock_checkpoint import minimum_mock_checkpoint

            # We remove all unessential metadata for testing purposes
            metadata = minimum_mock_checkpoint(metadata)

        if args.yaml:
            print(yaml.dump(metadata, indent=2, sort_keys=True), file=file)
            return

        if args.json or True:
            if args.pytest:
                print(json.dumps(metadata, sort_keys=True), file=file)
            else:
                print(json.dumps(metadata, indent=4, sort_keys=True), file=file)
            return

    def get(self, args: Namespace) -> None:
        """Navigate and print the metadata via a dot-separated path.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        from pprint import pprint

        from anemoi.utils.checkpoints import load_metadata

        metadata = load_metadata(args.path, name=args.name)

        if args.get == ".":
            print("Metadata from root: ", list(metadata.keys()))
            return

        for key in args.get.split("."):
            if key == "":
                keys = list(metadata.keys())
                print(f"Metadata keys from {args.get[:-1]}: ", keys)
                return
            else:
                metadata = metadata[key]

        print(f"Metadata values for {args.get}: ", end="\n" if isinstance(metadata, (dict, list)) else "")
        if isinstance(metadata, dict):
            pprint(metadata, indent=2, compact=True)
        else:
            print(metadata)

    def load(self, args: Namespace) -> None:
        """Load metadata into the checkpoint from a specified file.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        from anemoi.utils.checkpoints import has_metadata
        from anemoi.utils.checkpoints import replace_metadata
        from anemoi.utils.checkpoints import save_metadata

        if args.input is None:
            raise ValueError("Please specify a value for --input")

        _, ext = os.path.splitext(args.input)
        if ext == ".json" or args.json:
            with open(args.input) as f:
                metadata = json.load(f)

        elif ext in (".yaml", ".yml") or args.yaml:
            with open(args.input) as f:
                metadata = yaml.safe_load(f)

        else:
            raise ValueError(f"Unknown file extension {ext}. Please specify --json or --yaml")

        if has_metadata(args.path, name=args.name):
            replace_metadata(args.path, metadata)
        else:
            save_metadata(args.path, metadata, name=args.name)

    def supporting_arrays(self, args: Namespace) -> None:
        """Print the supporting arrays from the metadata.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        from anemoi.utils.checkpoints import load_metadata

        _, supporting_arrays = load_metadata(args.path, supporting_arrays=True)

        for name, array in supporting_arrays.items():
            print(f"{name}: shape={array.shape} dtype={array.dtype}")


command = Metadata



# (C) Copyright 2024-2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any

import earthkit.data as ekd
import tqdm
from anemoi.transform.fields import new_field_from_numpy
from anemoi.transform.fields import new_fieldlist_from_list

from anemoi.inference.context import Context

from ..processor import Processor
from . import pre_processor_registry

LOG = logging.getLogger(__name__)


@pre_processor_registry.register("no_missing_values")
class NoMissingValues(Processor):
    """Replace NaNs with mean."""

    def __init__(self, context: Context, **kwargs: Any) -> None:
        """Initialize the NoMissingValues processor.

        Parameters
        ----------
        context : Context
            The context in which the processor operates.
        kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context)

    def process(self, fields: ekd.FieldList) -> ekd.FieldList:
        """Process the fields to replace NaNs with the mean value.

        Parameters
        ----------
        fields : list
            List of fields to process.

        Returns
        -------
        list
            List of processed fields with NaNs replaced by the mean value.
        """
        result = []
        for field in tqdm.tqdm(fields):
            import numpy as np

            data = field.to_numpy()

            mean_value = np.nanmean(data)

            data = np.where(np.isnan(data), mean_value, data)
            result.append(new_field_from_numpy(data, template=field))

        return new_fieldlist_from_list(result)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import earthkit.data as ekd
from earthkit.data.utils.dates import to_datetime

from anemoi.inference.context import Context
from anemoi.inference.types import DataRequest
from anemoi.inference.types import Date
from anemoi.inference.types import State

from . import input_registry
from .grib import GribInput
from .mars import postproc

LOG = logging.getLogger(__name__)


def retrieve(
    requests: List[DataRequest],
    grid: Optional[Union[str, List[float]]],
    area: Optional[List[float]],
    dataset: Union[str, Dict[str, Any]],
    **kwargs: Any,
) -> ekd.FieldList:
    """Retrieve data from CDS.

    Parameters
    ----------
    requests : List[Dict[str, Any]]
        List of request dictionaries.
    grid : Optional[Union[str, List[float]]]
        Grid specification.
    area : Optional[List[float]]
        Area specification.
    dataset : Union[str, Dict[str, Any]]
        Dataset to use.
    **kwargs : Any
        Additional keyword arguments.

    Returns
    -------
    Any
        Retrieved data.
    """

    def _(r: DataRequest) -> str:
        mars = r.copy()
        for k, v in r.items():
            if isinstance(v, (list, tuple)):
                mars[k] = "/".join(str(x) for x in v)
            else:
                mars[k] = str(v)

        return ",".join(f"{k}={v}" for k, v in mars.items())

    pproc = postproc(grid, area)

    result = ekd.from_source("empty")
    for r in requests:
        if isinstance(dataset, str):
            d = dataset
        elif isinstance(dataset, dict):
            # Get dataset from intersection of keys between request and dataset dict
            search_dataset = dataset.copy()
            while isinstance(search_dataset, dict):
                keys = set(r.keys()).intersection(set(search_dataset.keys()))
                if len(keys) == 0:
                    raise KeyError(
                        f"While searching for dataset, could not find any valid key in dictionary: {r.keys()}, {search_dataset}"
                    )
                key = list(keys)[0]
                if r[key] not in search_dataset[key]:
                    if "*" in search_dataset[key]:
                        search_dataset = search_dataset[key]["*"]
                        continue

                    raise KeyError(
                        f"Dataset dictionary does not contain key {r[key]!r} in {key!r}: {dict(search_dataset[key])}."
                    )
                search_dataset = search_dataset[key][r[key]]

            d = search_dataset

        r.update(pproc)
        r.update(kwargs)

        LOG.debug("%s", _(r))

        result += ekd.from_source("cds", d, r)

    return result


@input_registry.register("cds")
class CDSInput(GribInput):
    """Get input fields from CDS."""

    trace_name = "cds"

    def __init__(
        self, context: Context, *, dataset: Union[str, Dict[str, Any]], namer: Optional[Any] = None, **kwargs: Any
    ) -> None:
        """Initialize the CDSInput.

        Parameters
        ----------
        context : Context
            The context in which the input is used.
        dataset : Union[str, Dict[str, Any]]
            The dataset to use.
        namer : Optional[Any]
            Optional namer for the input.
        **kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context, namer=namer)

        self.variables = self.checkpoint.variables_from_input(include_forcings=False)
        self.dataset = dataset
        self.kwargs = kwargs

    def create_input_state(self, *, date: Optional[Date]) -> State:
        """Create the input state for the given date.

        Parameters
        ----------
        date : Optional[Date]
            The date for which to create the input state.

        Returns
        -------
        State
            The created input state.
        """
        if date is None:
            date = to_datetime(-1)
            LOG.warning("CDSInput: `date` parameter not provided, using yesterday's date: %s", date)

        date = to_datetime(date)

        return self._create_input_state(
            self.retrieve(
                self.variables,
                [date + h for h in self.checkpoint.lagged],
            ),
            variables=self.variables,
            date=date,
        )

    def retrieve(self, variables: List[str], dates: List[Date]) -> Any:
        """Retrieve data for the given variables and dates.

        Parameters
        ----------
        variables : List[str]
            List of variables to retrieve.
        dates : List[Date]
            List of dates for which to retrieve data.

        Returns
        -------
        Any
            Retrieved data.
        """

        requests = self.checkpoint.mars_requests(
            variables=variables,
            dates=dates,
            use_grib_paramid=self.context.use_grib_paramid,
            patch_request=self.context.patch_data_request,
        )

        if not requests:
            raise ValueError("No requests for %s (%s)" % (variables, dates))

        return retrieve(
            requests, self.checkpoint.grid, self.checkpoint.area, dataset=self.dataset, expver="0001", **self.kwargs
        )

    def load_forcings_state(self, *, variables: List[str], dates: List[Date], current_state: State) -> State:
        """Load the forcings state for the given variables and dates.

        Parameters
        ----------
        variables : List[str]
            The list of variables for which to load the forcings state.
        dates : List[Date]
            The list of dates for which to load the forcings state.
        current_state : State
            The current state to be updated with the loaded forcings state.

        Returns
        -------
        Any
            The loaded forcings state.
        """
        return self._load_forcings_state(
            self.retrieve(variables, dates), variables=variables, dates=dates, current_state=current_state
        )



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from typing import Any

from anemoi.utils.registry import Registry

from anemoi.inference.config import Configuration

runner_registry = Registry(__name__)


def create_runner(config: Configuration, **kwargs: Any) -> Any:
    """Create a runner instance based on the given configuration.

    Parameters
    ----------
    config : Configuration
        The configuration for the runner.
    kwargs : dict
        Additional arguments for the runner.

    Returns
    -------
    Any
        The created runner instance.
    """
    return runner_registry.create(config.runner, config, **kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from anemoi.inference.input import Input
from anemoi.inference.output import Output
from anemoi.inference.processor import Processor
from anemoi.inference.types import IntArray

if TYPE_CHECKING:
    from .checkpoint import Checkpoint
    from .forcings import Forcings


LOG = logging.getLogger(__name__)


class Context(ABC):
    """Represents the context of the inference."""

    allow_nans = (None,)  # can be True of False
    use_grib_paramid = False
    verbosity = 0
    development_hacks: Dict[str, Any] = {}  # For testing purposes, don't use in production

    # Some runners will set these values, which can be queried by Output objects,
    # but may remain as None

    reference_date = None
    time_step = None
    lead_time = None
    output_frequency: Optional[int] = None
    write_initial_state: bool = True

    ##################################################################

    @property
    @abstractmethod
    def checkpoint(self) -> "Checkpoint":
        """Returns the checkpoint used for the inference."""
        pass

    ##################################################################
    # The methods below are not marked as abstract because they are
    # not required to be implemented by subclasses. However, they are
    # expected to be implemented by subclasses when relevant.
    # For example, when running the `SimpleRunner`, the user is
    # expected to provide the forcings directly as input to the runner.
    ##################################################################

    def create_input(self) -> Input:
        """Creates an input object for the inference.

        Returns
        -------
        Input
            The input object for the inference.
        """
        raise NotImplementedError()

    def create_output(self) -> Output:
        """Creates an output object for the inference.

        Returns
        -------
        Output
            The output object for the inference.
        """
        raise NotImplementedError()

    def create_constant_computed_forcings(self, variables: List[str], mask: IntArray) -> List["Forcings"]:
        """Creates constant computed forcings.

        Parameters
        ----------
        variables : List[str]
            List of variable names.
        mask : IntArray
            Mask array.

        Returns
        -------
        List[Forcings]
            List of constant computed forcings.
        """
        raise NotImplementedError(f"{self.__class__.__name__}.create_constant_computed_forcings")

    def create_constant_coupled_forcings(self, variables: List[str], mask: IntArray) -> List["Forcings"]:
        """Creates constant coupled forcings.

        Parameters
        ----------
        variables : List[str]
            List of variable names.
        mask : IntArray
            Mask array.

        Returns
        -------
        List[Forcings]
            List of constant coupled forcings.
        """
        raise NotImplementedError(f"{self.__class__.__name__}.create_constant_coupled_forcings")

    def create_dynamic_computed_forcings(self, variables: List[str], mask: IntArray) -> List["Forcings"]:
        """Creates dynamic computed forcings.

        Parameters
        ----------
        variables : List[str]
            List of variable names.
        mask : IntArray
            Mask array.

        Returns
        -------
        List[Forcings]
            List of dynamic computed forcings.
        """
        raise NotImplementedError(f"{self.__class__.__name__}.create_dynamic_computed_forcings")

    def create_dynamic_coupled_forcings(self, variables: List[str], mask: IntArray) -> List["Forcings"]:
        """Creates dynamic coupled forcings.

        Parameters
        ----------
        variables : List[str]
            List of variable names.
        mask : IntArray
            Mask array.

        Returns
        -------
        List[Forcings]
            List of dynamic coupled forcings.
        """
        raise NotImplementedError(f"{self.__class__.__name__}.create_dynamic_coupled_forcings")

    def create_pre_processors(self) -> List[Processor]:
        """Creates a list of pre-processors.

        Returns
        -------
        List[Processor]
            List of pre-processors.
        """
        return []

    def create_post_processors(self) -> List[Processor]:
        """Creates a list of post-processors.

        Returns
        -------
        List[Processor]
            List of post-processors.
        """
        return []



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import json
from argparse import ArgumentParser
from argparse import Namespace
from typing import Any
from typing import Callable

from ..checkpoint import Checkpoint
from . import Command


class InspectCmd(Command):
    """Inspect the contents of a checkpoint file."""

    def add_arguments(self, command_parser: ArgumentParser) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : ArgumentParser
            The argument parser to which the arguments will be added.
        """
        command_parser.add_argument("path", help="Path to the checkpoint.")
        command_parser.add_argument(
            "--validate", action="store_true", help="Validate the current virtual environment against the checkpoint"
        )

    def run(self, args: Namespace) -> None:
        """Run the inspect command.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        c = Checkpoint(args.path)

        if args.validate:
            c.validate_environment()
            return

        def _(f: Callable[[], Any]) -> Any:
            """Wrapper function to handle exceptions.

            Parameters
            ----------
            f : Callable
                The function to be called.

            Returns
            -------
            Any
                The result of the function call or the exception message.
            """
            try:
                return f()
            except Exception as e:
                return str(e)

        for name in sorted(dir(c)):

            if name.startswith("_"):
                continue

            print(name, ":")
            print("  ", json.dumps(_(lambda: getattr(c, name)), indent=4, default=str))
            print()


command = InspectCmd



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from argparse import ArgumentParser
from argparse import Namespace

from anemoi.utils.grib import shortname_to_paramid

from ..checkpoint import Checkpoint
from . import Command
from .retrieve import checkpoint_to_requests


class RequestCmd(Command):
    """MARS request utility."""

    def add_arguments(self, command_parser: ArgumentParser) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : ArgumentParser
            The argument parser to which the arguments will be added.
        """
        command_parser.description = self.__doc__
        command_parser.add_argument("--mars", action="store_true", help="Print the MARS request.")
        command_parser.add_argument("--use-grib-paramid", action="store_true", help="Use paramId instead of param.")
        command_parser.add_argument("path", help="Path to the checkpoint.")

    def run(self, args: Namespace) -> None:
        """Run the request command.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        c = Checkpoint(args.path)
        for r in checkpoint_to_requests(c, date=-1, use_grib_paramid=args.use_grib_paramid):
            if args.mars:
                req = ["retrieve,target=data"]
                for k, v in r.items():

                    if args.use_grib_paramid and k == "param":
                        if not isinstance(v, (list, tuple)):
                            v = [v]
                        v = [shortname_to_paramid(x) for x in v]

                    if isinstance(v, (list, tuple)):
                        v = "/".join([str(x) for x in v])
                    req.append(f"{k}={v}")
                r = ",".join(req)
            print(r)


command = RequestCmd



import datetime

import numpy as np

from anemoi.inference.runners.simple import SimpleRunner

# Create a runner with the checkpoint file
runner = SimpleRunner("checkpoint.ckpt")

# Select a starting date
date = datetime.datetime(2024, 10, 25)

# Assuming that the initial conditions requires two
# dates, e.g. T0 and T-6

multi_step_input = 2

# Define the grid

latitudes = np.linspace(90, -90, 181)  # 1 degree resolution
longitudes = np.linspace(0, 359, 360)

number_of_points = len(latitudes) * len(longitudes)
latitudes, longitudes = np.meshgrid(latitudes, longitudes)

# Create the initial state

input_state = {
    "date": date,
    "latitudes": latitudes,
    "longitudes": longitudes,
    "fields": {
        "2t": np.random.rand(multi_step_input, number_of_points),
        "msl": np.random.rand(multi_step_input, number_of_points),
        "z_500": np.random.rand(multi_step_input, number_of_points),
        ...: ...,
    },
}

# Run the model

for state in runner.run(input_state=input_state, lead_time=240):
    # This is the date of the new state
    print("New state:", state["date"])

    # This is value of a field for that date
    print("Forecasted 2t:", state["fields"]["2t"])



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from typing import Any
from typing import Dict
from typing import Protocol


class MetadataProtocol(Protocol):
    """Protocol for metadata objects. This will keep `mypy` happy."""

    _metadata: Dict[str, Any]
    _supporting_arrays: Dict[str, Any]
    grid: str
    variables: Dict[str, Any]



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Any
from typing import List
from typing import Optional

from anemoi.inference.types import Date
from anemoi.inference.types import State

if TYPE_CHECKING:
    from anemoi.inference.context import Context

LOG = logging.getLogger(__name__)

# TODO: only one method is need: `load_data`.
# The other methods can be implemenneted concreetly
# using the `load_data` method.


class Input(ABC):
    """Abstract base class for input handling."""

    trace_name = "????"  # Override in subclass

    def __init__(self, context: "Context"):
        """Initialize the Input object.

        Parameters
        ----------
        context : Context
            The context for the input.
        """
        self.context = context
        self.checkpoint = context.checkpoint

    def __repr__(self) -> str:
        """Return a string representation of the Input object.

        Returns
        -------
        str
            The string representation of the Input object.
        """
        return f"{self.__class__.__name__}()"

    @abstractmethod
    def create_input_state(self, *, date: Optional[Date]) -> State:
        """Create the input state dictionary.

        Parameters
        ----------
        date : Optional[Date]
            The date for which to create the input state.

        Returns
        -------
        State
            The input state dictionary.
        """
        pass

    @abstractmethod
    def load_forcings_state(self, *, variables: List[str], dates: List[Date], current_state: State) -> State:
        """Load forcings (constant and dynamic).

        Parameters
        ----------
        variables : List[str]
            The list of variables to load.
        dates : List[Date]
            The list of dates for which to load the forcings.
        current_state : State
            The current state of the model.

        Returns
        -------
        State
            The updated state with the loaded forcings.
        """
        pass

    def input_variables(self) -> List[str]:
        """Return the list of input variables.

        Returns
        -------
        List[str]
            The list of input variables.
        """
        return list(self.checkpoint.variable_to_input_tensor_index.keys())

    def set_private_attributes(self, state: State, value: Any) -> None:
        """Provide a way to a subclass to set private attributes in the state
        dictionary, that may be needed by the output object.

        Parameters
        ----------
        state : State
            The state dictionary.
        value : Any
            The value to set.
        """
        pass



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from typing import List
from typing import Tuple

from anemoi.inference.types import IntArray

from ..forcings import ComputedForcings
from ..forcings import Forcings
from ..runner import Runner
from . import runner_registry

LOG = logging.getLogger(__name__)


@runner_registry.register("plugin")
class PluginRunner(Runner):
    """A runner implementing the ai-models plugin API."""

    def __init__(self, checkpoint: str, *, device: str):
        """Initialize the PluginRunner.

        Parameters
        ----------
        checkpoint : str
            The checkpoint for the runner.
        device : str
            The device to run the model on.
        """
        super().__init__(checkpoint, device=device)

    # Compatibility with the ai_models API

    @property
    def param_sfc(self) -> List[str]:
        """Get surface parameters."""
        params, _ = self.checkpoint.mars_by_levtype("sfc")
        return sorted(params)

    @property
    def param_level_pl(self) -> Tuple[List[str], List[int]]:
        """Get pressure level parameters and levels."""
        params, levels = self.checkpoint.mars_by_levtype("pl")
        return sorted(params), sorted(levels)

    @property
    def param_level_ml(self) -> Tuple[List[str], List[int]]:
        """Get model level parameters and levels."""
        params, levels = self.checkpoint.mars_by_levtype("ml")
        return sorted(params), sorted(levels)

    @property
    def lagged(self) -> List[datetime.timedelta]:
        """Get lagged times in hours."""
        return self.checkpoint.lagged

    def create_constant_computed_forcings(self, variables: List[str], mask: IntArray) -> List[Forcings]:
        """Create constant computed forcings.

        Parameters
        ----------
        variables : List[str]
            The variables for the computed forcings.
        mask : IntArray
            The mask for the computed forcings.

        Returns
        -------
        List[Forcings]
            The constant computed forcings.
        """
        result = ComputedForcings(self, variables, mask)
        LOG.info("Constant computed forcing: %s", result)
        return [result]

    def create_dynamic_computed_forcings(self, variables: List[str], mask: IntArray) -> List[Forcings]:
        """Create dynamic computed forcings.

        Parameters
        ----------
        variables : list
            The variables for the computed forcings.
        mask : IntArray
            The mask for the computed forcings.

        Returns
        -------
        List[Forcings]
            The dynamic computed forcings.
        """
        result = ComputedForcings(self, variables, mask)
        LOG.info("Dynamic computed forcing: %s", result)
        return [result]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import logging
from typing import Any
from typing import List
from typing import Optional

import earthkit.data as ekd
import numpy as np

from ..types import Date
from ..types import State
from . import input_registry
from .grib import GribInput

LOG = logging.getLogger(__name__)


@input_registry.register("fdb")
class FDBInput(GribInput):
    """Get input fields from FDB."""

    trace_name = "fdb"

    def __init__(
        self,
        context,
        *,
        namer=None,
        fdb_config: dict | None = None,
        fdb_userconfig: dict | None = None,
        **kwargs: dict[str, Any],
    ):
        """Initialise the FDB input.

        Parameters
        ----------
        context : dict
            The context runner.
        namer : optional
            The namer to use for the input.
        fdb_config : dict, optional
            The FDB config to use.
        fdb_userconfig : dict, optional
            The FDB userconfig to use.
        kwargs : dict, optional
            Additional keyword arguments for the request to FDB.
        """
        super().__init__(context, namer=namer)
        self.kwargs = kwargs
        self.configs = {"config": fdb_config, "userconfig": fdb_userconfig}
        # NOTE: this is a temporary workaround for #191 thus not documented
        self.param_id_map = kwargs.pop("param_id_map", {})
        self.variables = self.checkpoint.variables_from_input(include_forcings=False)

    def create_input_state(self, *, date: Optional[Date]) -> State:
        date = np.datetime64(date).astype(datetime.datetime)
        dates = [date + h for h in self.checkpoint.lagged]
        ds = self.retrieve(variables=self.variables, dates=dates)
        res = self._create_input_state(ds, variables=None, date=date)
        return res

    def load_forcings_state(self, *, variables: List[str], dates: List[Date], current_state: State) -> State:
        ds = self.retrieve(variables=variables, dates=dates)
        return self._load_forcings_state(ds, variables=variables, dates=dates, current_state=current_state)

    def retrieve(self, variables: List[str], dates: List[Date]) -> Any:
        requests = self.checkpoint.mars_requests(
            variables=variables,
            dates=dates,
            use_grib_paramid=self.context.use_grib_paramid,
            patch_request=self.context.patch_data_request,
        )
        requests = [self.kwargs | r for r in requests]
        # NOTE: this is a temporary workaround for #191
        for request in requests:
            request["param"] = [self.param_id_map.get(p, p) for p in request["param"]]
        sources = [ekd.from_source("fdb", request, stream=False, **self.configs) for request in requests]
        ds = ekd.from_source("multi", sources)
        return ds



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import datetime
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from anemoi.inference.config import Configuration
from anemoi.inference.types import Date

LOG = logging.getLogger(__name__)


class CoupleConfiguration(Configuration):
    """Configuration class for the couple runner."""

    description: Optional[str] = None

    date: Optional[Date] = None
    """The starting date for the forecast. If not provided, the date will depend on the selected Input object. If a string, it is parsed by :func:`anemoi.utils.dates.as_datetime`."""

    lead_time: Optional[Tuple[str, int, datetime.timedelta]] = None
    """The lead time for the forecast. This can be a string, an integer or a timedelta object.
    If an integer, it represents a number of hours. Otherwise, it is parsed by :func:`anemoi.utils.dates.as_timedelta`.
    """

    name: Optional[str] = None
    """Used by prepml."""

    transport: str
    couplings: List[Dict[str, List[str]]]
    tasks: Dict[str, Any]

    env: Dict[str, Any] = {}
    """Environment variables to set before running the model. This may be useful to control some packages
    such as `eccodes`. In certain cases, the variables mey be set too late, if the package for which they are intended
    is already loaded when the runner is configured.
    """
